import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class InterRow:
    def __init__(self):
        self.map_timedelta = {
            "days": "D",
            "hours": "H",
            "minutes": "m",
            "seconds": "s",
        }

    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# InterRow")

        # group_by, field, new_column_name, periods, fillna, fillna_value, fillna_value_timedelta
        groupby: list = settings["group_by"] if ("group_by" in settings and settings["group_by"] != []) else []
        field: str = settings["field"] if ("field" in settings and settings["field"] != "") else None
        new_column_name: str = (
            settings["new_column_name"]
            if ("new_column_name" in settings and settings["new_column_name"] != "")
            else "new_column"
        )

        periods: int = settings["periods"] if (("periods" in settings and settings["periods"] != "") or settings["periods"] == 0) else 0

        inter_row_type: str = (
            settings["inter_row_type"]
            if ("inter_row_type" in settings and settings["inter_row_type"] != "")
            else None
        )
        fillna: bool = (
            settings["fillna"] if ("fillna" in settings) else False
        )

        fillna_value : str = (
            settings["fillna_value"]
            if ("fillna_value" in settings and settings["fillna_value"] != "")
            else None
        )

        fillna_value_timedelta = (
            settings["fillna_value_timedelta"]
            if ("fillna_value_timedelta" in settings and settings["fillna_value_timedelta"] != {})
            else {}
        )

        if not field or (fillna and (not fillna_value and not fillna_value_timedelta)):
            if not field:
                msg = "(inter_row) Debes seleccionar al menos una columna para aplicar la función inter_row"
            if fillna and (not fillna_value and not fillna_value_timedelta):
                msg = "(inter_row) Si seleccionas la función 'fillna' debes especificar el valor para los nulos"

            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        try:
            query = ""
            if groupby:
                if fillna:
                    if fillna_value:
                        query = f"df.groupby(by=groupby)[[field]].{inter_row_type}(periods).fillna(fillna_value)"
                        df[new_column_name] = eval(query)
                    else:
                        query = f"df.groupby(by=groupby)[[field]].{inter_row_type}(periods).fillna(self.timedelta_parse(fillna_value_timedelta))"
                        df[new_column_name] = eval(query)
                else:
                    query = f"df.groupby(by=groupby)[[field]].{inter_row_type}(periods)"
                    df[new_column_name] = eval(query)
            else:
                if fillna:
                    if fillna_value:
                        query = f"df[[field]].{inter_row_type}(periods).fillna(fillna_value)"
                        df[new_column_name] = eval(query)
                    else:
                        query = f"df[[field]].{inter_row_type}(periods).fillna(self.timedelta_parse(fillna_value_timedelta))"
                        df[new_column_name] = eval(query)
                else:
                    query = f"df[field].{inter_row_type}(periods)"
                    df[new_column_name] = eval(query)

            script.append(f"df[{new_column_name}] = {query}")

        except Exception as e:
            msg = "(inter_row) Exception: " + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}

    def timedelta_parse(self, timedelta_data):
        return " ".join([str(value) + " " + key for key, value in timedelta_data.items()])



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class ValueCounts:
    def exec(self, flow_id: str, node_key: str, pin: dict, settings: dict):
        script = []
        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# VALUE_COUNTS")
        if "field" not in settings:
            msg = "(value_counts) Debes seleccionar un columna para el campo Field"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        field: str = settings["field"]
        ascending: bool = (
            settings["ascending"]
            if ("ascending" in settings and settings["ascending"] is not None)
            else True
        )
        drop_na: bool = (
            settings["drop_na"]
            if ("drop_na" in settings and settings["drop_na"] is not None)
            else False
        )
        # normalize: bool = (
        #     settings["normalize"]
        #     if ("normalize" in settings and settings["normalize"] is not None)
        #     else False
        # )
        # print(field, ascending, drop_na, normalize)

        try:
            df_pct = df.value_counts(subset=field, normalize=True, ascending=ascending, dropna=drop_na)
            df_pct = df_pct.reset_index(name="value_pct")
            
            df_count = df.value_counts(subset=field, normalize=False, ascending=ascending, dropna=drop_na)
            df_count = df_count.reset_index(name="value_count")
            
            df = pd.merge(df_count, df_pct)
            
            script.append("\n# pct data")
            script.append("df_pct = df.value_counts(subset=[{}], normalize=True, ascending={}, dropna={})".format(field, ascending, drop_na))
            script.append("df_pct = df_pct.reset_index(name='value_pct')")
            
            script.append("\n# count data")
            script.append("df_count = df.value_counts(subset=[{}], normalize=False, ascending={}, dropna={})".format(field, ascending, drop_na))
            script.append("df_count = df_count.reset_index(name='value_count')")
            
            script.append("\n# merge data")
            script.append("df = pd.merge(df_count, df_pct)")
            
        except Exception as e:
            msg = "(value_counts) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Sort:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        # sufix_in = settings['port_In']

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# SORT")
        setting_list = list(map(lambda x: (x["field"], int(x["ascending"])), settings["items"]))
        columns, order = zip(*setting_list)

        try:
            df = df.sort_values(by=list(columns), ascending=list(order))
            script.append(
                "df_{} = df.sort_values(by=list({}), ascending=list({}))".format(
                    node_key, columns, order
                )
            )
        except Exception as e:
            msg = "(sort) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import pandas as pd

from evalml.automl.automl_search import AutoMLSearch
from vtarget.handlers.event_handler import event_handler


def find(arr: list, prop: str, value: any):
    for x in arr:
        if x[prop] == value:
            return x


class AutoML_Cache:
    def __init__(self):
        self.df: pd.DataFrame = None
        self.target: str = None
        self.problem_type: str = None
        self.models: list = []
        self.current_model_name: str = None
        self.partition_data: dict = {}
        self.automl_search: AutoMLSearch = None

    def reset(self):
        pass

    def find_model(self, model_name: str) -> dict:
        return find(self.models, "name", model_name)

    def get_df(self):
        return self.df

    def set_models(self, models: list):
        self.models = models
        event_handler.emit_queue.put(
            {
                "name": "automl.set_models",
                "data": json.dumps(models, default=str),
            }
        )

    def prepend_model(self, model: dict):
        self.models.insert(0, model)

    def append_model(self, model: dict):
        self.models.append(model)

    def is_a_model(self, model_name: str):
        model = self.find_model(model_name)
        return model is not None

    def get_current_model_name(self):
        return self.current_model_name

    def set_current_model_name(self, name: str):
        self.current_model_name = name

    def update_model(self, props: dict = {}, model_name: str = None):
        if model_name is None:
            model_name = self.get_current_model_name()
        model = self.find_model(model_name)
        if model is not None:
            model.update(props)
            event_handler.emit_queue.put(
                {
                    "name": "automl.update_model",
                    "data": json.dumps(model, default=str),
                }
            )

    # def update_models(self, models: list = []):
    #     for m in models:

    def get_partition_data(self) -> dict:
        return self.partition_data

    def set_partition_data(self, data: dict = {}):
        # self.partition_data.update(data)
        self.partition_data = data

    def set_automl_search(self, automl: AutoMLSearch):
        self.automl_search = automl

    def get_automl_search(self) -> AutoMLSearch:
        return self.automl_search


automl_cache = AutoML_Cache()



class Log_Handler:
    def __init__(self):
        self.log = []
        self.__append_handler = None

    def on_append(self, handler):
        self.__append_handler = handler

    def append_handler(self, log):
        if self.__append_handler is not None:
            self.__append_handler(log)

    def append(self, log, use_handler=True):
        self.log.append(log)
        if use_handler:
            self.append_handler(log)


log_handler = Log_Handler()



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class DatetimeExtract:
    def __init__(self):
        pass

    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# Datetime Extract")

        items: list[str] = ( settings["items"] if ("items" in settings and settings["items"] is not None) else None)

        if items:
            for item in items:
                # column_to_convert, new_column_name, to_extract
                column_to_convert: str = ( item["column_to_convert"] if ("column_to_convert" in item and item["column_to_convert"] != '') else '')
                to_extract: str = ( item["to_extract"] if ("to_extract" in item and item["to_extract"] != '') else '')
                new_column_name: str = item["new_column_name"] if "new_column_name" in item and item["new_column_name"] != '' else to_extract.replace("()", "")

                msg: str = None
                if not to_extract: msg = "(datetime_extract) Debes seleccionar al menos un campo a obtener para aplicar la función DateTime Extract"
                if not column_to_convert: msg = "(datetime_extract) Debes seleccionar al menos una columna para aplicar la función DateTime Extract"
                if msg: return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

                try:
                    # Forma 1
                    df[new_column_name] = eval("df[column_to_convert].dt.{}".format(to_extract))
                    script.append(f'df["{new_column_name}"] = df[column_to_convert].dt.{to_extract}')

                except Exception as e:
                    msg = "(datetime_extract) Exception:" + str(e)
                    return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

            cache_handler.update_node(
                flow_id,
                node_key,
                {
                    "pout": {"Out": df},
                    "config": json.dumps(settings, sort_keys=True),
                    "script": script,
                },
            )

            bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
            script_handler.script += script
            return {"Out": df}
        else:
            return bug_handler.default_on_error(flow_id, node_key, "(datetime_extract) Debes seleccionar al menos una columna para aplicar la función DateTime Extract", console_level="error")



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Cumsum:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# CUMSUM")
        # print(settings)
        groupby: list = settings["groupby"] if "groupby" in settings else []
        axis: str = settings["axis"] if "axis" in settings else None
        cumcount: bool = settings["cumcount"] if "cumcount" in settings else False
        cumsum: bool = settings["cumsum"] if "cumsum" in settings else False
        cumpct: bool = settings["cumpct"] if "cumpct" in settings else False
        pct: bool = settings["pct"] if "pct" in settings else False

        if not axis:
            msg = "(cumsum) Debes seleccionar una columna para aplicar la sumarización"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        prefix = axis[:3]
        script.append(f"prefix = '{prefix}'")

        if not (cumcount or cumsum or cumpct or pct):
            msg = "(cumsum) No se seleccionó ninguna agregación para la columna seleccionada"
            return bug_handler.default_on_error(
                flow_id, node_key, msg, console_level="warn", bug_level="warning", success=True
            )

        try:
            if groupby:
                df_obj = df.groupby(groupby)
                script.append(f"df_obj = df.groupby({groupby})['{axis}']")
                if cumcount:
                    df[prefix + "_cumcount"] = df_obj.cumcount() + 1
                    script.append(f"df[prefix+'_cumcount'] = df_obj.cumcount()+1")
                if cumsum:
                    df[prefix + "_cumsum"] = df_obj[axis].apply(lambda x: x.cumsum())
                    script.append(f"df[prefix+'_cumsum'] = df_obj.apply(lambda x: x.cumsum())")
                if pct:
                    df[prefix + "_pct"] = df_obj[axis].apply(lambda x: x / x.sum())
                    script.append(f"df[prefix+'_pct'] = df_obj.apply(lambda x: x/x.sum())")
                if cumpct:
                    df[prefix + "_cumpct"] = df_obj[axis].apply(lambda x: (x / x.sum()).cumsum())
                    script.append(
                        f"df[prefix+'_cumpct'] = df_obj.apply(lambda x: (x/x.sum()).cumsum())"
                    )
            else:
                if cumcount:
                    df[prefix + "_cumcount"] = range(1, 1 + len(df))
                    script.append(f"df[prefix+'_cumcount'] = range(1, 1 + len(df))")
                if cumsum:
                    df[prefix + "_cumsum"] = df[axis].cumsum()
                    script.append(f"df[prefix+'_cumsum'] = df['{axis}'].cumsum()")
                if pct:
                    df[prefix + "_pct"] = df[axis] / df[axis].sum()
                    script.append(f"df[prefix+'_pct'] = df['{axis}']/df[axis].sum()")
                if cumpct:
                    df[prefix + "_cumpct"] = df[axis].cumsum() / df[axis].sum()
                    script.append(
                        f"df[prefix+'_cumpct'] = df['{axis}'].cumsum()/df['{axis}'].sum()"
                    )

        except Exception as e:
            msg = "(cumsum) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Column:
    def __init__(self):
        self.script = []
        # self.sufix_in = ''

    def exec(self, flow_id, node_key, pin, settings):
        # self.sufix_in = settings['port_In']

        df = pin["In"].copy()
        self.script.append("\n# COLUMN")

        df, _, rename_cols = self.select_types_and_fields(flow_id, node_key, df, settings["items"])

        # Si hay alguna columna para renombrar en las seleccionadas
        if rename_cols:
            try:
                df = df.rename(columns=rename_cols)
            except Exception as e:
                msg = "(column) Error al renombrar, exception:" + str(e)
                return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

            self.script.append("\n# RENAME")
            self.script.append("df = df.rename(columns={})".format(rename_cols))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": self.script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += self.script.copy()
        self.script = []
        # return df, {'dtypes': dtypes}
        # print(df.dtypes)
        return {"Out": df}  # , {'dtypes': dtypes}

    # Retorna el df con las columnas seleccionadas y el tipo de dato
    def select_types_and_fields(self, flow_id, node_key, df, dtypes):
        # https://pbpython.com/pandas_dtypes.html

        # print('-----> dtypes: ', dtypes)
        # obtengo solo los campos seleccionados de la lista total de campos (selected==True)
        selected = dict(
            filter(
                lambda x: True if "selected" in x[1] and x[1]["selected"] else False,
                dtypes.items(),
            )
        )
        # print('-----> selected:', selected)

        # Se maneja la posibilidad de que ya no existan columnas que previamente fueron creadas
        available_cols = []
        removed_cols = []
        for field, _ in selected.items():
            if field in df.columns:
                available_cols.append(field)
            else:
                removed_cols.append(field)
                del dtypes[field]  # dado que no existe la eliminamos de dtypes
                msg = "(column) columa {} no existe en el df".format(field)
                bug_handler.default_on_error(
                    flow_id, node_key, msg, console_level="warn", bug_level="warning", success=True
                )

        # Remuevo las columnas que ya no existen
        for del_key in removed_cols:
            del selected[del_key]

        # mantiene solamente columnas existentes y seleccionadas
        df = df[available_cols]
        # print('-----> available_cols:', available_cols)
        # print('-----> df.columns:', df.columns)
        # print('-----> dtypes: ', dtypes)

        self.script.append("df = df[{}]".format(available_cols))
        self.script.append("\n# DATA TYPES")

        rename_cols = {}
        for field, _ in selected.items():
            # Genero el diccionario para el renombrado de variables
            if "rename" in dtypes[field] and dtypes[field]["rename"]:
                rename_cols[field] = dtypes[field]["rename"]

        # Ordeno las columnas segun el orden que se le dio en la interfaz
        order_cols = list(dict(sorted(selected.items(), key=lambda item: item[1]["order"])).keys())
        return df[order_cols], dtypes, rename_cols

    # Cambia el tipo de dato y maneja los errores que podrían salir en el intento
    def select_change_col_dtype(self, flow_id, node_key, df, field, dtype):
        try:
            # df = df.fillna('')
            df[field] = df[field].astype(dtype)
            # df.loc[:, [field]] = df[field].astype(dtype)
        except Exception as e:
            msg = "(column) No fue posible transformar un tipo de dato. Exception:" + str(e)
            bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            return df, False
        else:
            dtype_ = dtype if isinstance(dtype, str) else dtype.__name__
            self.script.append("df['{0}'] = df['{0}'].astype('{1}')".format(field, dtype_))
            return df, True



import gc
import pandas as pd
import json

from vtarget.dataprep.nodes.code import Code
from vtarget.dataprep.nodes.column import Column
from vtarget.dataprep.nodes.concat import Concat
from vtarget.dataprep.nodes.cross_join import CrossJoin
from vtarget.dataprep.nodes.cumsum import Cumsum
from vtarget.dataprep.nodes.data_cleansing import DataCleansing
from vtarget.dataprep.nodes.database import Database
from vtarget.dataprep.nodes.database_write import DatabaseWrite
from vtarget.dataprep.nodes.datetime_extract import DatetimeExtract
from vtarget.dataprep.nodes.datetime_formatter import DatetimeFormatter
from vtarget.dataprep.nodes.describe import Describe
from vtarget.dataprep.nodes.df_maker import DfMaker
from vtarget.dataprep.nodes.drop_duplicates import DropDuplicates
from vtarget.dataprep.nodes.dtype import Dtype
from vtarget.dataprep.nodes.email import Email
from vtarget.dataprep.nodes.excel import ExcelOutput
from vtarget.dataprep.nodes.filter import Filter
from vtarget.dataprep.nodes.formula import Formula
from vtarget.dataprep.nodes.groupby import Groupby
from vtarget.dataprep.nodes.input_data import InputData
from vtarget.dataprep.nodes.inter_row import InterRow
from vtarget.dataprep.nodes.isin import IsIn
from vtarget.dataprep.nodes.melt import Melt
from vtarget.dataprep.nodes.merge import Merge
from vtarget.dataprep.nodes.output import Output
from vtarget.dataprep.nodes.pivot import Pivot
from vtarget.dataprep.nodes.shape import Shape
from vtarget.dataprep.nodes.sort import Sort
from vtarget.dataprep.nodes.switch import Switch
from vtarget.dataprep.nodes.unique import Unique
from vtarget.dataprep.nodes.v_output import VOutput
from vtarget.dataprep.nodes.value_counts import ValueCounts
from vtarget.handlers.cache_handler import cache_handler
from vtarget.utils.utilities import utilities


class Pipeline:
    def __init__(self):
        self.decimal_round = False
        self.nodes_instances = {
            "Input_Data": InputData(),
            "Database": Database(),
            "Database_Write": DatabaseWrite(),
            "Sort": Sort(),
            "Filter": Filter(),
            "Formula": Formula(),
            "Merge": Merge(),
            "Group_By": Groupby(),
            "Cross_Join": CrossJoin(),
            "Concat": Concat(),
            "Pivot": Pivot(),
            "Shape": Shape(),
            "Melt": Melt(),
            "Output_Data": Output(),
            "Code": Code(),
            "Value_Counts": ValueCounts(),
            "Describe": Describe(),
            "Isin": IsIn(),
            "Cumsum": Cumsum(),
            "V_Output": VOutput(),
            "Inter_Row": InterRow(),
            "Unique": Unique(),
            "Drop_Duplicates": DropDuplicates(),
            "Data_Cleansing": DataCleansing(),
            "Datetime_Formatter": DatetimeFormatter(),
            "Datetime_Extract": DatetimeExtract(),
            "Switch": Switch(),
            "Select": Dtype(),
            "Dtype": Dtype(),
            "Column": Column(),
            "Excel": ExcelOutput(),
            "Email": Email(),
            "DF_Maker": DfMaker(),
            "Source": Code(),
        }

    def exec(self, flow_id: str, node: dict, input_port: dict):
        # print(node['key'])

        dict_pout: dict = self.nodes_instances[node["type"]].exec(
            flow_id,
            node["key"],
            input_port,
            node["meta"]["config"] if "config" in node["meta"] else {},
        )

        if "STDOUT" in dict_pout:
            node["meta"]["STDOUT"] = dict_pout["STDOUT"]

        cached_node = cache_handler.get_node(flow_id, node["key"])
        node["meta"]["script"] = cached_node["script"] if "script" in cached_node else []
        
        # agregar ports_config a caché
        if "ports_config" in node["meta"]:
            cache_handler.update_node(flow_id, node["key"], { "ports_config": json.dumps(node["meta"]["ports_config"], sort_keys=True)})

        for port_name in dict(node["meta"]["ports_map"]["pout"]).keys():
            port_config: dict = utilities.get_table_config(node["meta"], port_name)
            port_df: pd.DataFrame = dict_pout[port_name]
            
            node["meta"]["ports_map"]["pout"][port_name]["head"] = utilities.get_head_of_df_as_list(port_df, port_config, flow_id, node["key"], port_name)
            #TODO: revisar si la referencia de dict_pout se actualiza al modifcar cache por el sort_by del puerto
            node["meta"]["ports_map"]["pout"][port_name]["rows"] = port_df.shape[0]
            node["meta"]["ports_map"]["pout"][port_name]["cols"] = port_df.shape[1]
            node["meta"]["ports_map"]["pout"][port_name]["dtypes"] = utilities.get_dtypes_of_df(port_df)
            # node['meta']['ports_map']['pout'][port_name]['summary'] = utilities.viz_summary(port_df)
            # node['meta']['ports_map']['pout'][port_name]['describe'] = utilities.get_central_tendency_measures(port_df)
            node["meta"]["ports_map"]["pout"][port_name]["summary"] = {}
            node["meta"]["ports_map"]["pout"][port_name]["describe"] = {}
            node["meta"]["readed_from_cache"] = False
            
        del dict_pout
        gc.collect()
        return node


import base64

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad, unpad

# https://medium.com/@sachadehe/encrypt-decrypt-data-between-python-3-and-javascript-true-aes-algorithm-7c4e2fa3a9ff


class Encrypter:
    def encrypt(self, data: str, key: str, iv: str = "VTVTVTVTVTVTVTVT"):
        iv = iv.encode("utf-8")  # 16 char for AES128
        data = pad(data.encode(), 16)
        cipher = AES.new(key.encode("utf-8"), AES.MODE_CBC, iv)
        return base64.b64encode(cipher.encrypt(data))

    def decrypt(self, enc: str, key: str, iv: str = "VTVTVTVTVTVTVTVT"):
        iv = iv.encode("utf-8")  # 16 char for AES128
        enc = base64.b64decode(enc)
        cipher = AES.new(key.encode("utf-8"), AES.MODE_CBC, iv)
        b = unpad(cipher.decrypt(enc), 16)
        return b.decode("utf-8")


encrypter = Encrypter()



import datetime
import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class DatetimeFormatter:
    # def __init__(self):
    # self.pattern = re.compile(
    #     r"day|dd|Month|Mon|MM|yyyy|yy|hh|HH|mm|ss|,|\.|-|:|;|/|\||\s|de|las|a"
    # )
    # self.map_datetime = {
    #     "dy": "%j",
    #     "day": "%A",
    #     "dd": "%d",
    #     "Month": "%B",
    #     "Mon": "%b",
    #     "MM": "%m",
    #     "yy": "%y",
    #     "yyyy": "%Y",
    #     "HH": "%I",
    #     "hh": "%H",
    #     "mm": "%M",
    #     "ss": "%S",
    #     ",": ",",
    #     ".": ".",
    #     "-": "-",
    #     ":": ":",
    #     ";": ";",
    #     "/": "/",
    #     "|": "|",
    #     " ": " ",
    #     "de": "de",
    #     "las": "las",
    #     "a": "a",
    # }

    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# Datetime Formatter")

        # column_to_convert, new_column_name, pattern, custom_pattern
        column_to_convert: str = (
            settings["column_to_convert"]
            if "column_to_convert" in settings and settings["column_to_convert"] != None
            else None
        )
        new_column_name: str = (
            settings["new_column_name"]
            if "new_column_name" in settings and settings["new_column_name"] != ""
            else "new_column_name"
        )
        pattern: str = (
            settings["pattern"] if "pattern" in settings and settings["pattern"] != None else None
        )
        custom_pattern: str = (
            settings["custom_pattern"]
            if "custom_pattern" in settings and settings["custom_pattern"] != None
            else None
        )

        if not column_to_convert:
            msg = "(datetime_formatter) Debes seleccionar al menos una columna para aplicar la función DateTime Formatter"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        if custom_pattern:
            if custom_pattern[0] not in ["'", '"'] or custom_pattern[-1] not in ["'", '"']:
                msg = "(datetime_formatter) El patrón personalizado debe venir entre comillas"
                return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
            pattern = custom_pattern[1:-1]
        else:
            if not pattern:
                msg = "(datetime_formatter) Debes seleccionar al menos un formato para aplicar la función DateTime Formatter"
                return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
            # pattern = pattern[1:-1]

        try:
            script.append(
                f"""df["{new_column_name}"] = df.apply(lambda x: process_datetime(x["{column_to_convert}"], "{pattern}"), axis=1)"""
            )

            df[new_column_name] = df.apply(
                lambda x: self.process_datetime(x[column_to_convert], pattern, script), axis=1
            )

        except Exception as e:
            msg = "(datetime_formatter) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}

    def process_datetime(self, _datetime, pattern, script):
        # Hay que ver por qué ocurren estos 2 casos
        # Cuando es string entra como None
        if isinstance(_datetime, float):
            return None
        # Cuando es datetime entra como NaT
        if pd.isnull(_datetime):
            return None
        if not _datetime:
            return None

        result = _datetime
        # Es string
        if isinstance(_datetime, str):
            result = datetime.datetime.strptime(_datetime.lower(), pattern)
            if len(script) == 2:
                script.append(
                    f"""def process_datetime(_datetime, pattern): \n\treturn datetime.datetime.strptime(_datetime.lower(), pattern)"""
                )
        else:
            result = _datetime.strftime(pattern)
            if len(script) == 2:
                script.append(
                    f"""def process_datetime(_datetime, pattern): \n\treturn _datetime.strftime(pattern)"""
                )

        return result

    # def validate_custom_format(self, custom_pattern):
    #     new_pattern = ""
    #     matches = self.pattern.finditer(custom_pattern)
    #     for match in matches:
    #         pattern = self.get_datetime_pattern(match.group())
    #         if not pattern:
    #             return None
    #         else:
    #             new_pattern += pattern
    #         # print("Coincidencia:", match.group())
    #         # print("Posición inicial:", match.start())
    #         # print("Posición final:", match.end())
    #     return new_pattern

    # def get_datetime_pattern(self, pattern):
    #     return self.map_datetime[pattern] if pattern in self.map_datetime else None



import asyncio
import queue
import threading


class EventHandler:
    def __init__(self):
        self.emit_queue = queue.Queue()
        self.running = True
        self.__event_callback = None

    def on_event(self, callback):
        self.__event_callback = callback

    def event_callback(self, payload):
        if self.__event_callback is not None:
            self.__event_callback(payload)

    def run(self):
        print("Iniciando hilo envio de eventos")
        while self.running:
            payload = self.emit_queue.get(True)
            self.event_callback(payload)

    def init(self):
        asyncio.run(self.run())


event_handler = EventHandler()
worker = threading.Thread(target=event_handler.init)
worker.daemon = True
worker.start()



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Pivot:
    def exec(self, flow_id: str, node_key: str, pin: dict, settings: dict):
        script = []
        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# PIVOT")
        # print(settings)

        # Definición de las funciones de agregación
        def concat(x, sep):
            return sep.join(x)

        # print(settings.keys())
        validation = list(
            set(["col_group", "col_header", "col_value", "agg_method"]) - set(settings.keys())
        )
        if len(validation) != 0:
            msg = (
                "(pivot_table): Debes completar todos los campos en la configuración. Campos faltantes: "
                + str(validation)
            )
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        # Obtengo las configuraciones
        col_group: list[str] = settings["col_group"] if "col_group" in settings else []
        col_header: str = settings["col_header"] if "col_header" in settings else ""
        col_value: str = settings["col_value"] if "col_value" in settings else ""
        agg_method: list[str] = settings["agg_method"] if "agg_method" in settings else []
        separator: str = settings["separator"] if "separator" in settings else ","

        agg_fn = []
        for am in agg_method:
            # if am == 'Concatenate': agg_fn.append(lambda x: concatenate(x, separator))
            if am in ["Concatenate", "Concat"]:

                def lmbd(x):
                    return concat(x, separator)

                lmbd.__name__ = "concat"
                agg_fn.append(lmbd)
            elif am == "Count (Without Nulls)":
                agg_fn.append(lambda x: len(x.dropna().unique()))
            elif am == "Count (With Nulls)":
                agg_fn.append("count")
            elif am == "Average":
                agg_fn.append("mean")
            else:
                agg_fn.append(am.lower())
        # print(agg_fn)
        # pivotea la tabla utilizando los métodos de agregación seleccionados
        try:
            # df = pd.pivot_table(df, index=groupby, columns=header, values=values, aggfunc=agg_fn, margins=True).reset_index()
            df = pd.pivot_table(
                df, index=col_group, columns=col_header, values=col_value, aggfunc=agg_fn
            ).reset_index()
        except Exception as e:
            msg = "(pivot_table) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        # print(df.reset_index())
        df.columns = ["_".join(list(map(str, x))) if x[1] else x[0] for x in df.columns]
        script.append(
            "df = pd.pivot_table(df, index={}, columns={}, values={}, aggfunc={}).reset_index()".format(
                col_group, col_header, col_value, agg_fn
            )
        )
        script.append(
            'df.columns = ["_".join(list(map(str, x))) if x[1] else x[0] for x in df.columns ]'
        )

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import gc
from typing import Dict


class Cache_Handler:
    # class Cache_Handler(threading.Thread):
    cache: Dict[str, Dict]

    def __init__(self):
        # super().__init__()
        self.cache = dict()
        # self.daemon = True
        # self.queue = queue.Queue()
        # self.start()

    def reset(self, flow_id: str):
        if flow_id in self.cache:
            self.cache[flow_id].clear()
            del self.cache[flow_id]
            gc.collect()

    def get_node(self, flow_id: str, node_key: str) -> dict:
        if flow_id in self.cache:
            if node_key in self.cache[flow_id]:
                return self.cache[flow_id][node_key]
        return {}

    def update_node(self, flow_id: str, node_key: str, new_props: dict = {}):
        if flow_id in self.cache:
            if node_key not in self.cache[flow_id]:
                self.cache[flow_id][node_key] = dict()

            self.cache[flow_id][node_key].update(new_props)

    def delete_node(self, flow_id: str, node_key: str) -> bool:
        if flow_id in self.cache:
            if node_key in self.cache[flow_id]:
                del self.cache[flow_id][node_key]
                return True
        return False

    # def dump(self, flow_id: str, node_key: str) -> None:
    #     self.queue.put_nowait(
    #         (
    #             flow_id,
    #             node_key,
    #         )
    #     )

    # def run(self) -> None:
    #     while True:
    #         flow_id, node_key = self.queue.get()
    #         self.__dump(flow_id, node_key)

    # def load(self, flow_id: str, node_key: str) -> None:
    #     if flow_id in self.cache and node_key not in self.cache[flow_id]:
    #         if (
    #             os.path.exists("cache")
    #             and os.path.exists(f"cache/{flow_id}")
    #             and os.path.exists(f"cache/{flow_id}/{node_key}")
    #         ):
    #             try:
    #                 with open(f"cache/{flow_id}/{node_key}", "rb") as file:
    #                     self.cache[flow_id][node_key] = pickle.load(file)
    #             except:
    #                 traceback.print_exception(*sys.exc_info())

    # def __dump(self, flow_id: str, node_key: str) -> None:
    #     if flow_id in self.cache:
    #         try:
    #             if not os.path.exists("cache"):
    #                 os.mkdir("cache")
    #             if not os.path.exists(f"cache/{flow_id}"):
    #                 os.mkdir(f"cache/{flow_id}")
    #             with open(f"cache/{flow_id}/{node_key}", "wb") as file:
    #                 pickle.dump(self.cache[flow_id], file)
    #         except:
    #             import sys
    #             import traceback

    #             traceback.print_exception(*sys.exc_info())

    # def free(self, flow_id: str, node_key: str) -> None:
    #     if flow_id in self.cache and node_key in self.cache[flow_id]:
    #         self.cache[flow_id][node_key].clear()
    #         del self.cache[flow_id][node_key]
    #         gc.collect()


cache_handler = Cache_Handler()



import os
import uuid


def run_flow(path: str) -> None:
    import gc
    import json

    from vtarget.dataprep.builder import Builder
    from vtarget.handlers.bug_handler import bug_handler

    content = open(path, "r")
    data = json.loads(content.read())
    builder = Builder()
    builder.init_pipeline()
    flow_id = str(uuid.uuid4())
    builder.analyzer(
        data["model"],
        True,
        flow_id,
        os.path.basename(path),
        False,
        False,
    )
    del builder.pipeline
    gc.collect()
    return bug_handler.bug

if __name__ == '__main__':
    print(run_flow('C:\\Users\\aflor\\Downloads\\model_crec_proy_cat_nf.json'))


import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler
from vtarget.utils.utilities import utilities


class Dtype:
    def __init__(self):
        self.script = []
        # self.sufix_in = ''

    def exec(self, flow_id: str, node_key: str, pin: dict, settings: dict):
        # self.sufix_in = settings['port_In']

        df: pd.DataFrame = pin["In"].copy()
        self.script.append("\n# DTYPE")

        df, _, rename_cols = self.select_types_and_fields(flow_id, node_key, df, settings["items"])

        # Si hay alguna columna para renombrar en las seleccionadas
        if rename_cols:
            try:
                df = df.rename(columns=rename_cols)
            except Exception as e:
                msg = "(dtype) Error al renombrar Exception:" + str(e)
                return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

            self.script.append("\n# RENAME")
            self.script.append("df = df.rename(columns={})".format(rename_cols))

        # if node_key not in cache_handler.cache[flow_id]:
        #     cache_handler.cache[flow_id][node_key] = dict()
        # cache_handler.cache[flow_id][node_key]["pout"] = {"Out": df}
        # cache_handler.cache[flow_id][node_key]["config"] = json.dumps(settings, sort_keys=True)
        # cache_handler.cache[flow_id][node_key]["script"] = self.script
        cache_handler.update_node(
            flow_id, 
            node_key, 
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": self.script,
            }
        )
        
        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += self.script.copy()
        self.script = []
        # return df, {'dtypes': dtypes}
        # print(df.dtypes)
        return {"Out": df}  # , {'dtypes': dtypes}

    # Retorna el df con las columnas seleccionadas y el tipo de dato
    def select_types_and_fields(self, flow_id: str, node_key: str, df: pd.DataFrame, dtypes: dict):
        # https://pbpython.com/pandas_dtypes.html

        # print('-----> dtypes: ', dtypes)
        # obtengo solo los campos seleccionados de la lista total de campos (selected==True)
        
        # ? para nodos de tipo DTYPE usar todos los campos, sin importar si estan o no en la config
        if 'dtype' in node_key.lower(): #aas
            all_dtypes = utilities.get_dtypes_of_df(df) # ? todos los dtypes del Dataframe
            dtypes = all_dtypes | dtypes
            
        
        selected_dtypes = dict(
            filter(
                lambda x: True if "selected" in x[1] and x[1]["selected"] else False,
                dtypes.items(),
            )) if 'select' in node_key.lower() else dtypes
        
        # selected = dtypes.copy()
        # print("-----> selected:", selected)
        # print("-----> length:", len(selected.items()), len(dtypes.items()))

        # Se maneja la posibilidad de que ya no existan columnas que previamente fueron creadas
        available_cols = []
        removed_cols = []
        for field, x in selected_dtypes.items():
            if field in df.columns:
                available_cols.append(field)
            else:
                removed_cols.append(field)
                del dtypes[field]  # dado que no existe la eliminamos de dtypes
                msg = "(dtype) columa {} no existe en el df".format(field)
                bug_handler.default_on_error(
                    flow_id, node_key, msg, console_level="warn", bug_level="warning", success=True
                )

        # Remuevo las columnas que ya no existen
        for del_key in removed_cols:
            del selected_dtypes[del_key]

        # mantiene solamente columnas existentes y seleccionadas
        df = df[available_cols]
        # print('-----> available_cols:', available_cols)
        # print('-----> df.columns:', df.columns)
        # print('-----> dtypes: ', dtypes)

        self.script.append("df = df[{}]".format(available_cols))
        self.script.append("\n# DATA TYPES")

        rename_cols = {}
        for field, x in selected_dtypes.items():
            # Genero el diccionario para el renombrado de variables
            if "rename" in dtypes[field] and dtypes[field]["rename"]:
                rename_cols[field] = dtypes[field]["rename"]

            standard_dtypes = [
                "object",
                "bool",
                "category",
                "int8",
                "int16",
                "int32",
                "int64",
                "float16",
                "float32",
                "float64",
            ]
            if x["dtype"] in standard_dtypes:
                # si el campo ya es un numérico conocido no lo cambio
                if df[field].dtype == x["dtype"]:
                    continue
                df, status = self.select_change_col_dtype(
                    flow_id, node_key, df.copy(), field, x["dtype"]
                )
                if not status:  # si hubo un error lo vuelvo a texto
                    dtypes[field]["dtype"] = "object"

            elif x["dtype"] in ["datetime64[ns]"]:
                try:
                    df = df.copy()
                    df[field] = pd.to_datetime(df[field])
                except Exception as e:
                    msg = (
                        "(dtype) No fue posible transformar un el campo {} a {}. Exception:".format(
                            field, x["dtype"], str(e)
                        )
                    )
                    bug_handler.default_on_error(flow_id, node_key, msg, str(e))
                    # si hubo un error lo vuelvo a texto
                    dtypes[field]["dtype"] = "object"

            elif x["dtype"] in ["timedelta64[ns]"]:
                try:
                    df = df.copy()
                    df[field] = pd.to_timedelta(df[field])
                except Exception as e:
                    msg = (
                        "(dtype) No fue posible transformar un el campo {} a {}. Exception:".format(
                            field, x["dtype"], str(e)
                        )
                    )
                    bug_handler.default_on_error(flow_id, node_key, msg, str(e))
                    # si hubo un error lo vuelvo a texto
                    dtypes[field]["dtype"] = "object"
            else:
                msg = '(dtype) Tipo de dato "{}" desconocido, campo "{}" se mantiene como string '.format(
                    x["dtype"], field
                )
                bug_handler.default_on_error(
                    flow_id, node_key, msg, console_level="warn", bug_level="warning"
                )
                dtypes[field]["dtype"] = "object"

        # print(df.dtypes)
        # print(dtypes)
        # print(rename_cols)
        # print(' --------------- \n\n ')
        # Ordeno las columnas segun el orden que se le dio en la interfaz
        order_cols = list(dict(sorted(selected_dtypes.items(), key=lambda item: item[1]["order"])).keys())
        return df[order_cols], dtypes, rename_cols

    # Cambia el tipo de dato y maneja los errores que podrían salir en el intento
    def select_change_col_dtype(self, flow_id, node_key, df, field, dtype):
        try:
            # df = df.fillna('')
            df[field] = df[field].astype(dtype)
            # df.loc[:, [field]] = df[field].astype(dtype)
        except Exception as e:
            msg = "(dtype) No fue posible transformar un tipo de dato. Exception:" + str(e)
            bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            return df, False
        else:
            dtype_ = dtype if isinstance(dtype, str) else dtype.__name__
            self.script.append("df['{0}'] = df['{0}'].astype('{1}')".format(field, dtype_))
            return df, True



import json

import numpy as np
import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Groupby:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info')
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df = pin["In"].copy()
        # print('\n\n\n node_key: ', node_key, '\n\n\n')
        script.append("\n# GROUPBY")

        def percentile(n):
            def percentile_(x):
                return np.percentile(x, n)

            percentile_.__name__ = "q%s" % n
            return percentile_

        def count_distinct(x):
            return x.nunique()

        # count_distinct = lambda x: x.nunique()

        def count_null(x):
            return x.isnull().sum()

        def mode(x):
            # return x.value_counts().index[0]
            return x.value_counts().idxmax()

        def count_blank(x):
            return sum(x == "")

        def count_not_blank(x):
            return sum(x == "")

        # https://www.analyticsvidhya.com/blog/2020/03/groupby-pandas-aggregating-data-python/
        group_by_cols = settings["group_by"] if "group_by" in settings else []
        agg_cols = {}
        # pctl_dynamic = {}
        # pctl_replaces = []
        pctl_replaces2 = []
        rename_cols = {}
        # group_by_rename = {}

        # print(settings['group_by'])
        # print(settings['agg'])

        for a in settings["agg"]:
            # if a['action'] == 'group_by': # obtengo las columnas del groupby
            # 	group_by_cols.append(a['column'])
            # 	if 'rename' in a and a['rename']: # agrego los rename asociado a la columna de agrupación
            # 		rename_cols[a['column']] = a['rename']
            # else:
            action = a["action"]
            if action == "percentile":
                # print('----------------------------------------------------')
                # print(action, a['pctl_value'])
                fn_name = "quantile_{}".format(a["pctl_value"])
                # user_fn_name = 'p{}_{}'.format(a['pctl_value'], a['column'])
                # print(int(a['pctl_value'])/100)
                # pctl_dynamic[fn_name] = lambda x :  x.quantile(q=int(a['pctl_value'])/100)
                # pctl_dynamic[fn_name].__name__ = 'p{}'.format(a['pctl_value'])
                # exec('{} = lambda x :  x.quantile(q={})'.format(fn_name, int(a['pctl_value'])/100))
                # # locals()[fn_name] = lambda x :  x.quantile(q=int(a['pctl_value'])/100)
                # exec('{}.__name__ = "{}"'.format(fn_name, user_fn_name))
                # exec("a['action'] = quantile_{}".format(a['pctl_value']))
                # a['action'] = pctl_dynamic[fn_name]
                action = fn_name
                # pctl_replaces.append(fn_name)
                pctl_replaces2.append((fn_name, a["pctl_value"]))

            # luego agrupo las funciones de agregación
            if a["column"] not in agg_cols:
                agg_cols[a["column"]] = [action]
            else:
                # valido que no se agreguen agregaciones repetidas
                if action not in agg_cols[a["column"]]:
                    agg_cols[a["column"]].append(action)
            # Si viene una columna de agregación con un renombre desde la vista
            if "rename" in a and a["rename"]:
                # print(a['column'], '->', a['rename'])
                # creo el nombre compuesto entre la columna y la fn de agg
                current_name = a["column"] + "_" + action
                rename_cols[current_name] = a["rename"]
        # print(pctl_dynamic)
        agg_str = str(agg_cols)
        # for pr in pctl_replaces:
        for pr in pctl_replaces2:
            # agg_str = agg_str.replace("'{}'".format(pr), "pctl_dynamic['{}']".format(pr))
            agg_str = agg_str.replace("'{}'".format(pr[0]), "percentile({})".format(pr[1]))
        if "count_distinct" in agg_str:
            agg_str = agg_str.replace("'count_distinct'", "count_distinct")
        if "count_null" in agg_str:
            agg_str = agg_str.replace("'count_null'", "count_null")
        if "mode" in agg_str:
            agg_str = agg_str.replace("'mode'", "mode")

        # print('\n -------------------------------------------')
        # print(df)
        # # print(settings)
        # print('group_by_cols:', group_by_cols) # group_by_cols: ['segmento_1']
        # print('agg_cols: ', agg_cols) # agg_cols:  {'rut': ['count_distinct']}
        # print('agg_str: ', agg_str) # agg_str:  {'rut': [count_distinct]}
        # print('rename_cols: ', rename_cols) #
        # print('\n -------------------------------------------')

        grouped = pd.DataFrame()
        try:
            if group_by_cols:
                # print('df.groupby(group_by_cols).agg({}).reset_index()'.format(agg_str))
                # exec('grouped = df.groupby(group_by_cols).agg({}).reset_index()'.format(agg_str))
                grouped = eval("df.groupby(group_by_cols).agg({}).reset_index()".format(agg_str))
                # print('grouped.columns (raw):', grouped.columns)
                # Dado que las columnas vienen en un multiIndex, con esto reseteo el indice
                grouped.columns = ["_".join(x) if x[1] else x[0] for x in grouped.columns]
                # print('grouped.columns (renamed):', grouped.columns)
                script.append(
                    "grouped = df.groupby({}).agg({}).reset_index()".format(group_by_cols, agg_str)
                )
            else:
                # df = df.copy()
                grouped = eval("df.groupby(lambda _ : 1).agg({}).reset_index()".format(agg_str))
                # print('\n\n')
                # print(grouped.columns)
                # print(list(grouped.columns))
                grouped.columns = ["_".join(x) if x[1] else x[0] for x in grouped.columns]
                grouped.drop(columns=["index"], axis=1, inplace=True)
                script.append(
                    "grouped = df.groupby(lambda _ : 1).agg({}).reset_index()".format(agg_str)
                )
        except Exception as e:
            msg = "(groupby) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        try:
            grouped.rename(columns=rename_cols, inplace=True)
        except Exception as e:
            msg = "(groupby) No fue realizar el rename luego de la sumarización. Exception:" + str(
                e
            )
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": grouped},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script

        return {"Out": grouped}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Filter:
    def __init__(self):
        self.script = []
        # self.sufix_in = ''

    def exec(self, flow_id, node_key, pin, settings):
        # self.sufix_in = settings['port_In']

        if "rule_type" not in settings:
            msg = "(filter) No se ha creado ninguna una regla"
            bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
            return {"T": pd.DataFrame(), "F": pd.DataFrame()}

        # if "sensitive_case" in settings and settings["sensitive_case"]:
        #     settings["value"] = settings["value"].lower()

        df: pd.DataFrame = pin["In"].copy()
        df_T, df_F = pd.DataFrame(), pd.DataFrame()
        # print(settings)
        if settings["rule_type"] == "preconfigured":
            try:
                df_T, df_F = self.filter_preconfigured(flow_id, node_key, df, settings)
            except Exception as e:
                msg = "(filter) Exception:" + str(e)
                bug_handler.default_on_error(flow_id, node_key, msg, str(e))
                return {"T": pd.DataFrame(), "F": pd.DataFrame()}

        elif settings["rule_type"] == "code":
            df_T, df_F = self.filter_code(flow_id, node_key, df, settings["sentence"])
        else:
            msg = '(filter_code) No reconoce la regla "{}"'.format(settings["rule_type"])
            bug_handler.default_on_error(
                flow_id, node_key, msg, console_level="warn", bug_level="warning"
            )

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"T": df_T, "F": df_F},
                "config": json.dumps(settings, sort_keys=True),
                "script": self.script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += self.script.copy()
        self.script = []

        return {"T": df_T, "F": df_F}

    # Nodo: filter - Procesa el filtro de sentencia
    def filter_code(self, flow_id, node_key, df, sentence):
        self.script.append("\n# FILTER (CUSTOMIZED)")
        # self.script.append(f"df = df_{self.sufix_in}")
        str_rule = "df[{}]".format(sentence)
        # print(df.dtypes)
        # print(str_rule)
        try:
            df_T = eval(str_rule)
            # df_T = pd.eval(str_rule)
            self.script.append("df_T = {}".format(str_rule))
        except Exception as e:
            msg = "(filter_code) No fue posible procesar la condición. Exception:" + str(e)
            bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            return pd.DataFrame(), pd.DataFrame()
        else:
            # Obtengo el complemento (la parte negada de la condición)
            df_F = df[~df.index.isin(df_T.index)]
            self.script.append("df_F = df[~df.index.isin(df_T.index)]")
            return df_T, df_F

    # Nodo: filter - Procesa el filtro preconfigurado
    def filter_preconfigured(self, flow_id, node_key, df: pd.DataFrame, rule):
        df_T, df_F = pd.DataFrame(), pd.DataFrame()
        self.script.append("\n# FILTER (PRECONFIGURED)")
        # self.script.append(f"df = df_{self.sufix_in}")
        field: str = rule["field"]
        operator: str = rule["operator"]
        value: str = rule["value"] if "value" in rule else '""'
        sensitive_case: bool = (
            True
            if pd.api.types.is_string_dtype(df[field])
            and "sensitive_case" in rule
            and rule["sensitive_case"]
            else False
        )

        if field not in df.columns:
            msg = "(filter_code) No existe la columna {} en el dataframe de entrada".format(field)
            bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
            return pd.DataFrame(), pd.DataFrame()

        if pd.api.types.is_datetime64_any_dtype(df[field]):
            value = "'{}'".format(value)

        # print( '{} {} {}'.format(field, operator, value) )
        # print(df.dtypes)
        # is_numeric_dtype(df['B'])
        # Lógica para crear los filtros
        if operator == "=":  # Numeric, String
            if pd.api.types.is_string_dtype(df[field]):
                value = "'{}'".format(value)

            str_rule = "df[(df['{}'] == {})]".format(field, value)
            if not sensitive_case:
                str_rule = "df[(df['{}'].str.lower() == {}.lower())]".format(field, value)

            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "!=":  # Numeric, String
            if pd.api.types.is_string_dtype(df[field]):
                value = "'{}'".format(value)

            str_rule = "df[(df['{}'] != {})]".format(field, value)
            if not sensitive_case:
                str_rule = "df[(df['{}'].str.lower() != {}.lower())]".format(field, value)
            # print('\n\n', str_rule, '\n\n')
            # print(df.dtypes)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "<":  # Numeric
            str_rule = "df[(df['{}'] < {})]".format(field, value)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "<=":  # Numeric
            str_rule = "df[(df['{}'] <= {})]".format(field, value)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == ">":  # Numeric
            str_rule = "df[(df['{}'] > {})]".format(field, value)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == ">=":  # Numeric
            str_rule = "df[(df['{}'] >= {})]".format(field, value)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "is_null":  # Numeric, String
            df_T = df[df[field].isnull()]
            self.script.append("df_T = df[df['{}'].isnull()]".format(field))
        elif operator == "is_not_null":  # Numeric, String
            df_T = df[df[field].notnull()]
            self.script.append("df_T = df[df['{}'].notnull()]".format(field))
        elif operator == "contains":  # String
            df_T = df[df[field].str.contains(value, case=sensitive_case)]
            self.script.append(
                "df_T = df[df['{}'].str.contains('{}', case={})]".format(
                    field, value, sensitive_case
                )
            )
        elif operator == "does_not_contain":  # String
            df_T = df[~df[field].str.contains(value, case=sensitive_case)]
            self.script.append(
                "df_T = df[~df['{}'].str.contains('{}', case={})]".format(
                    field, value, sensitive_case
                )
            )
        elif operator == "is_empty":  # String
            str_rule = "df[(df['{}'] == " ")]".format(field)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "is_not_empty":  # String
            str_rule = "df[(df['{}'] != " ")]".format(field)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "is_true":  # True
            str_rule = "df[(df['{}'] == True)]".format(field)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        elif operator == "is_false":  # False
            str_rule = "df[(df['{}'] == False)]".format(field)
            df_T = eval(str_rule)
            self.script.append(f"df_T = {str_rule}")
        else:
            msg = '(filter_code) Operador no reconocido "{}"'.format(rule["operator"])
            bug_handler.default_on_error(
                flow_id, node_key, msg, console_level="error", bug_level="error"
            )

        # Obtengo el complemento (la parte negada de la condición)
        df_F = df[~df.index.isin(df_T.index)]
        self.script.append(f"df_F = df[~df.index.isin(df_T.index)]")
        # print(df.head())
        return df_T, df_F



import json

import numpy as np
import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class DfMaker:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        script.append("\n# DFMAKER")
        data: list = settings["data"] if "data" in settings and settings["data"] is not None else []

        df = pd.DataFrame()

        # TODO: agregar script
        try:
            columns = []
            rows = []
            if len(data):
                columns = [c["value"] or f"col_{idx+1}" for idx, c in enumerate(data[0])]
                if len(data) > 1:
                    rows = [[c["value"] for c in r] for r in data[1:]]

            df = pd.DataFrame(np.array(rows), columns=columns)

        except Exception as e:
            msg = "(df_maker) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json
import math

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class ExcelOutput:
    def exec(self, flow_id, node_key, pin, settings):
        script_handler.script.append("\n# EXCEL")
        output_name: str = (
            settings["name"] if "name" in settings and settings["name"] is not None else "output"
        )
        output_path: str = (
            settings["path"] if "path" in settings and settings["path"] is not None else ""
        )
        encoding: str = settings["encoding"] or "UTF-8"

        # print(columns, order)
        df1 = pin["In"].copy() if "In" in pin else pd.DataFrame()
        df2 = pin["In2"].copy() if "In2" in pin else pd.DataFrame()
        df3 = pin["In3"].copy() if "In3" in pin else pd.DataFrame()
        df4 = pin["In4"].copy() if "In4" in pin else pd.DataFrame()
        df5 = pin["In5"].copy() if "In5" in pin else pd.DataFrame()

        sheet_name_1: str = (
            settings["sheet1"]
            if "sheet1" in settings and settings["sheet1"] != ""
            else "sheet_1"
        )
        sheet_name_2: str = (
            settings["sheet2"]
            if "sheet2" in settings and settings["sheet2"] != ""
            else "sheet_2"
        )
        sheet_name_3: str = (
            settings["sheet3"]
            if "sheet3" in settings and settings["sheet3"] != ""
            else "sheet_3"
        )
        sheet_name_4: str = (
            settings["sheet4"]
            if "sheet4" in settings and settings["sheet4"] != ""
            else "sheet_4"
        )
        sheet_name_5: str = (
            settings["sheet5"]
            if "sheet5" in settings and settings["sheet5"] != ""
            else "sheet_5"
        )

        file_path: str = f"{output_path}\\{output_name}.xlsx"
        settings["file_path"] = file_path

        # TODO: Agregar script
        try:
            # create a excel writer object
            with pd.ExcelWriter(file_path) as writer:
                # use to_excel function and specify the sheet_name and index
                # to store the dataframe in specified sheet
                if len(df1) > 0:
                    self.pd_to_excel(df1, writer, sheet_name_1, encoding)
                    # df1.to_excel(writer, sheet_name=sheet_name_1, index=False, encoding=encoding)
                if len(df2) > 0:
                    self.pd_to_excel(df2, writer, sheet_name_2, encoding)
                    # df2.to_excel(writer, sheet_name=sheet_name_2, index=False, encoding=encoding)
                if len(df3) > 0:
                    self.pd_to_excel(df3, writer, sheet_name_3, encoding)
                    # df3.to_excel(writer, sheet_name=sheet_name_3, index=False, encoding=encoding)
                if len(df4) > 0:
                    self.pd_to_excel(df4, writer, sheet_name_4, encoding)
                    # df4.to_excel(writer, sheet_name=sheet_name_4, index=False, encoding=encoding)
                if len(df5) > 0:
                    self.pd_to_excel(df5, writer, sheet_name_5, encoding)
                    # df5.to_excel(writer, sheet_name=sheet_name_5, index=False, encoding=encoding)

        except Exception as e:
            msg = f"(excel) No fue posible generar el archivo excel: {str(e)}"
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        out = []
        sheet_names = [sheet_name_1, sheet_name_2, sheet_name_3, sheet_name_4, sheet_name_5]
        for idx, df in enumerate([df1, df2, df3, df4, df5]):
            if len(df):
                out.append(
                    {
                        "Name": sheet_names[idx],
                        "Columns": len(df.columns),
                        "Rows": len(df),
                        # "File_Path": file_path,
                    }
                )

        df = pd.DataFrame(out)

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script_handler.script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        return {"Out": df}

    def pd_to_excel(self, df: pd.DataFrame, writer: pd.ExcelWriter, sheet_name: str, encoding: str):
        # https://support.microsoft.com/en-us/office/excel-specifications-and-limits-1672b34d-7043-467e-8e27-269d656771c3?ui=en-us&rs=en-us&ad=us
        rows: int = len(df)
        rows_per_sheet = 1048576 - 1  # max rows by worksheet = 1048576
        if rows > rows_per_sheet:
            steps = math.ceil(rows / rows_per_sheet)
            start = 0
            end = rows_per_sheet
            for idx in range(steps):
                step_df = df.copy().iloc[start:end]
                step_df.to_excel(
                    writer, sheet_name=f"{sheet_name}_{idx + 1}", index=False, encoding=encoding
                )
                start = end
                end += rows_per_sheet
        else:
            df.to_excel(writer, sheet_name=sheet_name, index=False, encoding=encoding)



import math

import numpy as np
from sklearn import metrics as skmetrics


class CalcMetrics:
    def __init__(self):
        pass

    def MASE(self, training_series, testing_series, prediction_series):
        """
        Computes the MEAN-ABSOLUTE SCALED ERROR forcast error for univariate time series prediction.

        See "Another look at measures of forecast accuracy", Rob J Hyndman

        parameters:
                training_series: the series used to train the model, 1d numpy array
                testing_series: the test series to predict, 1d numpy array or float
                prediction_series: the prediction of testing_series, 1d numpy array (same size as testing_series) or float
                absolute: "squares" to use sum of squares and root the result, "absolute" to use absolute values.

        """
        # print "Needs to be tested."
        n = training_series.shape[0]
        d = np.abs(np.diff(training_series)).sum() / (n - 1)

        errors = np.abs(testing_series - prediction_series)
        return errors.mean() / d

    def main_ts_metrics(self, real, pred, train):
        R2 = skmetrics.r2_score(real, pred)
        MAPE = skmetrics.mean_absolute_percentage_error(real, pred)
        MAE = skmetrics.mean_absolute_error(real, pred)
        MSE = skmetrics.mean_squared_error(real, pred)

        return {
            "R2": R2,
            "MAPE": MAPE,
            "MAE": MAE,
            "MSE": MSE,
            "RMSE": math.sqrt(MSE),
            "MASE": self.MASE(train, real, pred) if len(train) else None,
        }



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Formula:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        # sufix_in = settings['port_In']

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# FORMULA")
        # script.append(f"df_{node_key}_Out = df_{sufix_in}")
        # script.append(f"df = df_{sufix_in}")

        for i, item in enumerate(settings["items"]):
            col_name: str = ""
            if not item["field"]:  # crea una columna nueva
                col_name = (
                    item["new_column_name"]
                    if "new_column_name" in item and item["new_column_name"]
                    else f"x_{i}"
                )
            else:  # Edita una existente
                col_name = item["field"]
            if "sentence" not in item:
                continue

            try:
                # pd.options.mode.chained_assignment = None
                # print('---------------------------------------------------------')
                # print(col_name)
                # print('---------------------------------------------------------')
                # print(f['sentence'])
                # print(df.dtypes)
                # df[col_name] = eval(f['sentence'])# lanza un warning de este modo
                df = df.copy()  # con esto evito los warning
                df.loc[:, col_name] = eval(item["sentence"])
                # print(df.dtypes)
                # print(parent_ports)
                # df_name = 'df' if parent_ports[0] == 'Out' else 'df_'+parent_ports[0] # genero el nombre del df dependiendo de la salida del puerto padre
                script.append("df.loc[:, '{}'] = {}".format(col_name, item["sentence"]))
                # s = f['sentence'].replace('df[', f'df_{sufix_in}[').replace('df.', f'df_{sufix_in}.')
                # script.append("df_{}_Out.loc[:, '{}'] = {}".format(node_key, col_name, f['sentence']))
                # df[:][col_name] = eval(f['sentence'])
                # [5]*df.shape[0] # esto cuando se intenta agregar 1 solo valor a una columna completa
                # print(df.dtypes)
            except Exception as e:
                msg = f"(formula) No fue posible procesar la fórmula para la columna {col_name}. Exception: {str(e)}"
                return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class DropDuplicates:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# Drop Duplicates")

        # columns, keep, only_subset

        columns: list[str] = (
            settings["columns"] if ("columns" in settings and settings["columns"] is not []) else []
        )
        keep: str = (
            settings["keep"] if ("keep" in settings and settings["keep"] is not None) else "first"
        )
        only_subset: bool = (
            settings["only_subset"]
            if ("only_subset" in settings and settings["only_subset"] is not None)
            else False
        )

        try:
            if len(columns) == 0:
                df = df.drop_duplicates(keep=keep)
                script.append(f"df = df.drop_duplicates(keep={keep})")
            else:
                df = df.drop_duplicates(subset=columns, keep=keep)
                script.append(f"df = df.drop_duplicates(subset={columns}, keep={keep})")

            if only_subset:
                df = df[columns]
                script.append(f"df = df[{columns}]")

        except Exception as e:
            msg = "(drop_duplicates) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Shape:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# SHAPE")
        script.append("df_shape = pd.DataFrame([df.shape], columns=['num_rows', 'num_columns'])")
        df = pd.DataFrame([df.shape], columns=["num_rows", "num_columns"])

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import gc
import json
import time

import pandas as pd

from vtarget.dataprep.pipeline import Pipeline
from vtarget.dataprep.types import NodeType
from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.event_handler import event_handler
from vtarget.handlers.log_handler import log_handler
from vtarget.handlers.script_handler import script_handler
from vtarget.utils.utilities import utilities


class Builder:
    def __init__(self):
        # self.compute_node_counter = 0
        self.nodes: dict = {}
        self.script: str = ""
        self.init_script()  # resetea el script
        self.cur_node_key: str = None
        # self.ultra = None

    def init_pipeline(self):
        self.pipeline = Pipeline()

    def init_script(self):
        # Headers del script
        script_handler.script.append("#!/usr/bin/env python")
        script_handler.script.append("# coding: utf-8\n")
        script_handler.script.append("import pandas as pd")
        script_handler.script.append("import numpy as np")

    def set_nodes(self, flow_id: str, data: dict):
        # node_data = {}
        self.nodes = (
            {}
        )  # diccionario con los nodos del flujo cuya clave es el atributo "key" del nodo
        for idx, nd in enumerate(data["nodeDataArray"]):
            if (
                not ("isGroup" in nd and nd["isGroup"])  # no es grupo
                and (nd["type"] not in ["Comment"])  # no es comment
                and (nd["category"] not in ["chart"])  # no es chart
            ):
                self.nodes[nd["key"]] = {
                    "idx": idx,
                    "name": nd["name"],
                    "type": nd["type"],
                    "key": nd["key"],
                    "loaded": False,
                    "output": None,
                    "parents": [],
                    "childs": [],
                    # "is_group": "isGroup" in nd and nd["isGroup"],
                    "skip": False,
                }

        # recorrer los links para obtener los padres e hijos por nodo
        for ld in data["linkDataArray"]:
            if "_chart" in ld["to"]:
                continue
            # Agrego los padres
            parent = {"from": ld["from"], "frompid": ld["frompid"], "topid": ld["topid"]}
            if not self.nodes[ld["to"]]["parents"]:
                self.nodes[ld["to"]]["parents"] = [parent]
            else:
                self.nodes[ld["to"]]["parents"].append(parent)

            # Agrego los hijos
            child = {
                "to": ld["to"],
                "topid": ld["topid"],
                "frompid": ld["frompid"],
                "to_id": self.nodes[ld["to"]]["idx"],
            }
            if not self.nodes[ld["from"]]["childs"]:
                self.nodes[ld["from"]]["childs"] = [child]
            else:
                self.nodes[ld["from"]]["childs"].append(child)

        # print(self.node_data)
        # print('Tamaño inicial self.node_data (bytes): ',sys.getsizeof(self.node_data))

    # Resetea todos los nodos hijos cuando cambia la configuración de un padre
    def reset_childs_recursive(self, flow_id: str, node_key: str, reseted_nodes: list):
        for child in self.nodes[node_key]["childs"]:
            if child["to"] in cache_handler.cache[flow_id]:
                del cache_handler.cache[flow_id][child["to"]]
                # print('-----------------> caché reseteada para nodo',childs['to'])
                reseted_nodes.append(child["to"])
            self.reset_childs_recursive(flow_id, child["to"], reseted_nodes)
        return reseted_nodes

    # Setea a True el parámetro para saltar (no-procesar) los nodos hijos cuando existe un error
    def set_skip_childs_recursive(self, flow_id: str, node_key: str, data: dict):
        for child in self.nodes[node_key]["childs"]:
            # habilita la bandera para saltarse el nodo
            self.nodes[child["to"]]["skip"] = True
            # resetea el df completo almacenado en RAM
            self.nodes[child["to"]]["output"] = None
            # resetea el nodo de gojs
            # data['nodeDataArray'][childs['to_id']]['meta']['ports_map']['pout'] = {}
            for pout_name in data["nodeDataArray"][child["to_id"]]["meta"]["ports_map"]["pout"]:
                data["nodeDataArray"][child["to_id"]]["meta"]["ports_map"]["pout"][pout_name][
                    "head"
                ] = []
                data["nodeDataArray"][child["to_id"]]["meta"]["ports_map"]["pout"][pout_name][
                    "rows"
                ] = 0
                data["nodeDataArray"][child["to_id"]]["meta"]["ports_map"]["pout"][pout_name][
                    "cols"
                ] = 0
                # data['nodeDataArray'][childs['to_id']]['meta']['ports_map']['pout'][pout_name]['dtypes'] = None
            data["nodeDataArray"][child["to_id"]]["meta"]["skipped"] = True
            print("=================> se saltará el nodo", child["to"])
            self.set_skip_childs_recursive(flow_id, child["to"], data)
        return True

    def analyzer(
        self,
        data: dict,
        reset_cache: bool,
        flow_id: str,
        flow_name: str,
        disable_all_write_nodes: bool,
        decimal_round,
    ):
        # Reseteo de variables singleton para cada ejecución
        log_handler.log = []
        bug_handler.bug = []
        script_handler.script = []
        self.pipeline.decimal_round = decimal_round

        # Si se corre el flujo reseteando la cache
        if reset_cache:
            msg = f'Caché reseteada para el flujo "{flow_name}"'
            bug_handler.default_on_error(
                flow_id, None, msg, console_level="trace", bug_level="info", success=True
            )
            cache_handler.reset(flow_id)

        # Incializa la caché para el nodo si es que aún no existe
        if flow_id not in cache_handler.cache:
            cache_handler.cache[flow_id] = dict()

        cache_nodes_keys = list(cache_handler.cache[flow_id].keys())
        msg = f"{len(cache_nodes_keys)} Nodos en caché"
        bug_handler.default_on_error(
            flow_id, None, msg, console_level="trace", bug_level="info", success=True
        )

        # Inicializa una copia local de los nodos y su metadata
        self.set_nodes(flow_id, data)

        num_nodes_ok: int = 0  # conteo del número de nodos procesados
        attemps: int = 0  # contador de intentos del while para procesar el flujo completo
        t1 = time.time()  # inicializa el tiempo total de procesamiento de los nodos
        completed_nodes: list = []

        # mientras no estén todos los nodos procesados
        while num_nodes_ok < len(self.nodes):
            # print(num_nodes_ok, "<", len(self.nodes))
            attemps += 1
            for node_key, node_data in self.nodes.items():
                # Salto los nodos que están cargado
                if self.nodes[node_key]["loaded"]:
                    continue

                # * Salto los nodos output data cuando el check de deshabilitar está activado
                if disable_all_write_nodes and (node_data["type"] in [NodeType.OUTPUTDATA.value, NodeType.EXCEL.value]):
                    msg = "Se omite escritura de archivo para nodo {}".format(node_data["key"])
                    bug_handler.default_on_error(
                        flow_id, None, msg, console_level="trace", bug_level="info", success=True
                    )

                    # para que no entre dos veces en caso que el while haga mas de un intento
                    self.nodes[node_key]["loaded"] = True
                    num_nodes_ok += 1
                    completed_nodes.append(node_data["key"])
                    continue

                # Omito procesar los nodos hijos de un nodo que tuvo un error
                if self.nodes[node_key]["skip"]:
                    # self.node_data[n_key]['loaded'] = True # para que no entre dos veces en caso que el while haga mas de un intento
                    bug_handler.console(
                        "nodo {} saltado".format(node_data["key"]), "trace", flow_id
                    )
                    num_nodes_ok += 1
                    completed_nodes.append(node_data["key"])
                    continue

                # --------------------------------------------------
                # Dertermina si el/los nodos padres están cargados
                # --------------------------------------------------
                is_loaded: bool = False
                input_port: dict = {}  # almancena los df del mapeo de los puertos de entrada
                # parent_ports = []
                # Si es una entrada de datos que no tiene padre, se puede procesar
                if not node_data["parents"]:
                    # Omite los nodos de entrada de datos (sin puertos de entrada), ya que por consecuencia no tiene registro de nodos padres
                    if node_data["type"] in [
                        NodeType.INPUTDATA.value,
                        NodeType.DFMAKER.value,
                        NodeType.DATABASE.value,
                        NodeType.SOURCE.value,
                    ]:
                        is_loaded = True
                    else:
                        msg = "Se omite nodo {} sin entrada padre".format(node_data["key"])
                        bug_handler.default_on_error(
                            flow_id,
                            None,
                            msg,
                            console_level="trace",
                            bug_level="info",
                            success=True,
                        )

                        # para que no entre dos veces en caso que el while haga mas de un intento
                        self.nodes[node_key]["loaded"] = True
                        num_nodes_ok += 1
                        completed_nodes.append(node_data["key"])
                        self.set_skip_childs_recursive(flow_id, node_key, data)
                        continue
                else:  # Si el nodo tiene padres
                    # Compruebo si todos los padres tienen sus salidas cargadas
                    for parent in node_data["parents"]:
                        is_loaded = True  # incializo la carga en verdadero
                        # print(parent['from'], parent['from'] in cache_handler.cache[flow_id], cache_handler.cache[flow_id].keys())
                        # Si está la salida en el nodo padre
                        if self.nodes[parent["from"]]["output"] != None:
                            # if parent['from'] in cache_handler.cache[flow_id]:

                            # input_port[parent["topid"]] = cache_handler.cache[flow_id][parent["from"]]["pout"][parent["frompid"]]
                            parent_key = parent["from"]
                            parent_port = parent["frompid"]
                            parent_cache = cache_handler.get_node(
                                flow_id, parent_key
                            )  # nodo desde caché
                            
                            if (
                                parent_cache
                                and "pout" in parent_cache
                                and parent_port in parent_cache["pout"]
                            ):
                                input_port[parent["topid"]] = parent_cache["pout"][parent_port]
                            # else:
                            #     # TODO: error

                            # Almaceno en el mapeo de entrada del nodo el df asociado a la salida del padre
                            # input_port[parent['topid']] = self.node_data[parent['from']]['output'][parent['frompid']]
                            # parent_ports.append((parent['frompid'], self.node_data[parent['from']]))
                        else:  # Si al menos un padre no está procesado, etonces termina la iteración
                            # print('\n\n\n\n', 'x'*100, n_data['key'], parent['topid'], '--', parent['from'] , parent['frompid'] , '\n\n\n\n')
                            is_loaded = False  # si al menos uno no está cargado
                            break
                    # if n_data['name'] == 'Code':
                    # 	is_loaded = True
                # --------------------------------------------------
                # Si el nodo se puede procesar inicia el procesamiento
                # --------------------------------------------------
                # print(f'is_loaded: {is_loaded} / nodo: ', n_key)
                if is_loaded:
                    node_idx: int = self.nodes[node_key]["idx"]
                    node_name: str = data["nodeDataArray"][node_idx]["name"]
                    node_key: str = data["nodeDataArray"][node_idx]["key"]
                    self.cur_node_key = node_key

                    # print(f'PROCESANDO NODO {node_key}')

                    # --------------------------------------------------
                    # Actualizo los dtypes de la entrada con el dataframe de la salida del nodo anterior
                    # --------------------------------------------------
                    # Omite nodos de entrada de datos (sin puertos de entrada)
                    if node_data["type"] not in [
                        NodeType.INPUTDATA.value,
                        NodeType.DFMAKER.value,
                        NodeType.DATABASE.value,
                        NodeType.SOURCE.value,
                    ]:
                        for _, pin_name in enumerate(
                            data["nodeDataArray"][node_idx]["meta"]["ports_map"]["pin"]
                        ):
                            if (
                                "dtypes"
                                in data["nodeDataArray"][node_idx]["meta"]["ports_map"]["pin"][
                                    pin_name
                                ]
                                and pin_name in input_port
                            ):
                                dtypes_list = data["nodeDataArray"][node_idx]["meta"]["ports_map"][
                                    "pin"
                                ][pin_name]["dtypes"]

                                u_dtypes = self.update_inputs_dtypes(
                                    node_name,
                                    node_key,
                                    dtypes_list,
                                    input_port[pin_name],
                                )
                                data["nodeDataArray"][node_idx]["meta"]["ports_map"]["pin"][
                                    pin_name
                                ]["dtypes"] = u_dtypes
                                # print(pin_index, 'fin')

                    # --------------------------------------------------
                    # Valida si el nodo actual está en caché y si su configuración sigue siendo la misma
                    # --------------------------------------------------
                    node_cache = cache_handler.get_node(flow_id, node_key)
                    if (
                        node_cache 
                        and 
                        ("config" in node_cache and node_cache["config"] == json.dumps(data["nodeDataArray"][node_idx]["meta"]["config"], sort_keys=True))
                        and
                        (
                            "ports_config" not in node_cache
                            or 
                            ("ports_config" in node_cache and node_cache["ports_config"] == json.dumps(data["nodeDataArray"][node_idx]["meta"]["ports_config"], sort_keys=True))
                        )
                    ):
                        # await socket_server.emit('dataprep.node_processing', {
                        #     'flow_id': flow_id,
                        #     'key': n_key,
                        # })
                        # bug_handler.console(f'Nodo "{node_key}" saltado, para flujo flow_id: "{flow_id}"', 'info', flow_id, False)
                        # reset_childs = False
                        data["nodeDataArray"][node_idx]["meta"]["readed_from_cache"] = True
                        script_handler.script += node_cache["script"]
                        data["nodeDataArray"][node_idx]["meta"]["processed"] = True
                        # await socket_server.emit('dataprep.node_processed', {
                        #     'flow_id': flow_id,
                        #     'key': node_key,
                        #     'node': json.dumps(data['nodeDataArray'][idx], default=str),
                        #     # 'log': log_handler.log, # TODO: Hacer el log del nodo correspondiente
                        # })
                    # --------------------------------------------------
                    # Procesa el nodo correspondiente y actualiza los valores de los df
                    # --------------------------------------------------
                    else:
                        bug_handler.console(f"PROCESANDO NODO {node_key}", "-", flow_id)
                        if (
                            node_cache
                            and 
                            (
                                ("config" in node_cache and node_cache["config"] != json.dumps(data["nodeDataArray"][node_idx]["meta"]["config"], sort_keys=True))
                                or
                                ("ports_config" in node_cache and node_cache["ports_config"] != json.dumps(data["nodeDataArray"][node_idx]["meta"]["ports_config"], sort_keys=True))
                            )
                        ):
                            reseted_nodes = self.reset_childs_recursive(
                                flow_id, node_key, [node_key]
                            )
                            event_handler.emit_queue.put(
                                {
                                    "name": "dataprep.reseted_nodes",
                                    "data": {
                                        "flow_id": flow_id,
                                        "reseted_nodes": reseted_nodes,
                                    },
                                }
                            )

                        # await socket_server.emit('dataprep.node_processing', {
                        #     'flow_id': flow_id,
                        #     'key': n_key,
                        # })
                        data["nodeDataArray"][node_idx] = self.pipeline.exec(
                            flow_id, data["nodeDataArray"][node_idx], input_port
                        )
                        data["nodeDataArray"][node_idx]["meta"]["processed"] = True
                        event_handler.emit_queue.put(
                            {
                                "name": "dataprep.node_processed",
                                "data": {
                                    "flow_id": flow_id,
                                    "key": node_key,
                                    "node": json.dumps(
                                        data["nodeDataArray"][node_idx], default=str
                                    ),
                                    # 'log': log_handler.log, # TODO: Hacer el log del nodo correspondiente
                                },
                            }
                        )

                    # self.node_data[n_key]['output'] = full_df # Almaceno los df del mapeo de salida
                    # Almaceno los df del mapeo de salida
                    self.nodes[node_key]["output"] = True
                    # Fijo la bandera mara para marcar que el nodo está procesado
                    self.nodes[node_key]["loaded"] = True
                    # Aumento el contador para saber que ese nodo fue analizado y salir del while
                    num_nodes_ok += 1
                    completed_nodes.append(node_data["key"])

                    # --------------------------------------------------
                    # Valida si el nodo actual se registró en el singleton de bug con level==error, de estarlo se saltan los hijos de esos nodos
                    # --------------------------------------------------
                    has_error = (
                        next(
                            (
                                x
                                for x in bug_handler.bug
                                if x["node_key"] == node_key and x["level"] == "error"
                            ),
                            None,
                        )
                        != None
                    )
                    data["nodeDataArray"][node_idx]["meta"]["skipped"] = False
                    data["nodeDataArray"][node_idx]["meta"]["has_error"] = False
                    if has_error:
                        print("\n\n\n----------- has_error", has_error)
                        cache_handler.delete_node(flow_id, node_key)
                        data["nodeDataArray"][node_idx]["meta"]["has_error"] = True
                        self.set_skip_childs_recursive(flow_id, node_key, data)

                    # # Si ese nodo fue ejecutado nuevamente por algun cambio en la config, reseteo todos los hijos dependientes del mismo
                    # if reset_childs:
                    # 	# print(self.node_data[n_key]['childs'])
                    # 	# print('key: ', node_key, n_key)
                    # 	self.reset_childs_recursive(flow_id, n_key)

        # --------------------------------------------------
        # Fin del while y for principal
        # --------------------------------------------------
        cache_nodes_keys = list(cache_handler.cache[flow_id].keys())
        msg = f"Se almacenaron {len(cache_nodes_keys)} nodos en caché"
        # bug_handler.console(msg, 'debug', flow_id)
        bug_handler.default_on_error(
            flow_id, None, msg, console_level="debug", bug_level="info", success=True
        )

        msg = "Flujo procesado en {} segundos ".format(round(time.time() - t1, 3))
        bug_handler.default_on_error(
            flow_id, None, msg, console_level="debug", bug_level="info", success=True
        )

        # bug_handler.console(f'**** BUG **** {bug_handler.bug}', 'debug', flow_id)

        # Una vez que acaba la ejecución del flujo se conforma el script
        self.script = "\n".join(script_handler.script)
        # print('\n\n script', len(script_handler.script))
        # print(self.script)
        # # FIXME: No se puede, es más lento que la escritura del pickle en disco
        # t = time.time()
        # self.ultra = UltraDict(
        #     cache_handler.cache[flow_id],
        #     name=flow_id,
        #     recurse=True,
        #     shared_lock=True,
        #     # auto_unlink=True,
        # )
        # print(f"dump {time.time()-t} segundos")
        del self.nodes
        gc.collect()

        return data

    # ---------------------------------------------------------------------------------------
    # Actualiza los dtypes utilizando el df que se está recibiendo de entrada
    # y compara con lo que se tenía en la configuración, manejando tanto la
    # creación de campos que antes no existían, como la eliminación de campos que fueron eliminados

    def update_inputs_dtypes(self, node_name, node_key, current_dtypes, input_df: pd.DataFrame):
        # if node_name == 'Merge':
        # print('\ncurrent_dtypes')
        # print(current_dtypes)
        # print('\ninput_df')
        # print(input_df)

        # print('\nIteration')
        res = input_df.dtypes.to_frame("dtypes")
        res = res["dtypes"].astype(str).reset_index()
        # print('res:', res)
        updated_dtypes = {}
        for i, x in res.iterrows():
            updated_dtypes[x["index"]] = {
                "dtype": x["dtypes"],
                "selected": True,
                "order": i,
            }

        return updated_dtypes
        """
        if not len(current_dtypes):
            return current_dtypes

        # Si hay que crear campos nuevos se agregarán al final
        try:
            max_order =  max(list(map(lambda x: x['order'], current_dtypes.values())))
        except Exception as e:
            print('Error (builder): ', e)
            bug_handler.append({'flow_id':flow_id, 'success': False, 'node_key': None, 'level': 'error',
                                        'msg': 'No fue obtener el max de la lista de dtypes', 'exception': str(e)})
            return current_dtypes

        for i,x in res.iterrows():
            if x['index'] in current_dtypes: # si el campo en la salida está en el pin de entrada, sólo actualizo el tipo de dato
                if node_name != 'Select': # los select, al permitir cambiar los datatypes no deben actualizarse
                    current_dtypes[x['index']]['dtype'] = x['dtypes']
            else: # si el campo no está, es porque se editó el flujo en algun punto intermedio y se debe crear el campo
                # del current_dtypes[x['index']]
                max_order += 1
                current_dtypes[x['index']] = {'dtype': x['dtypes'], 'selected': True, 'order': max_order}
                print('previamente no existía el campo "{}" en el nodo "{}", se agrega'.format(x['index'], node_name))
                bug_handler.append({'flow_id':flow_id, 'success': True, 'node_key': node_key, 'level': 'info',
                                        'msg': 'Campo "{}" no existía en el nodo "{}" previamente, se agrega'.format(x['index'], node_name), 'exception': ''})
                # print('Campo "{}" no existe en nodo "{}" será omitido'.format(x['index'], node_name))
        # print('\n\n\ncurrent_dtypes:\n')
        # print(current_dtypes)
        # Extraigo los campos que antes existían y ya no
        removed_fields = list(set(current_dtypes.keys()) - set(res['index'].tolist()))
        # Remuevo los campos que ya no existen
        # current_dtypes = dict(filter(lambda i: i[0] in res['index'].tolist(), current_dtypes.items()))
        for rf in removed_fields:
            to_remove = current_dtypes[rf]
            # print(rf, to_remove)
            bug_handler.append({'flow_id':flow_id, 'success': True, 'node_key': node_key, 'level': 'warning',
                                        'msg': 'Ya no existe el campo "{}" en el nodo "{}", se elimina de sus dtypes'.format(rf, node_name), 'exception': ''})
            del current_dtypes[rf]

        # print('\ncurrent_dtypes (modified):')
        # print(current_dtypes)
        return current_dtypes
        """

    def remove_flow_from_cache(self, key: str):
        if key in cache_handler.cache:
            del cache_handler.cache[key]

    def remove_nodes_from_cache(self, flow_id: str, node_keys: list[str]):
        removeds = []
        if flow_id in cache_handler.cache:
            for nodekey in node_keys:
                if nodekey in cache_handler.cache[flow_id]:
                    removeds.append(nodekey)
                    del cache_handler.cache[flow_id][nodekey]
        return removeds

    # Actualiza node['meta']['ports_map']['pout'][port_name]['summary']
    def load_detailed_view(self, flow_id: str, node_key: str, port_name: str):
        if flow_id in cache_handler.cache and node_key in cache_handler.cache[flow_id]:
            df = cache_handler.cache[flow_id][node_key]["pout"][port_name]
            return utilities.viz_summary(df)
        return {}

    # Actualiza node['meta']['ports_map']['pout'][port_name]['describe']
    def load_column_view(self, flow_id: str, node_key: str, port_name: str):
        if flow_id in cache_handler.cache and node_key in cache_handler.cache[flow_id]:
            df = cache_handler.cache[flow_id][node_key]["pout"][port_name]
            return utilities.get_central_tendency_measures(df)
        return {}

    def modify_node(self, flow_id: str, node_key: str, node: dict, port_name: str):
        cached_node = cache_handler.get_node(flow_id, node_key)
        if cached_node and "pout" in cached_node and port_name in cached_node["pout"]:
            port_df: pd.DataFrame = cached_node["pout"][port_name]
            port_config: dict = utilities.get_table_config(node["meta"], port_name)
            # for port_name in node["meta"]["ports_map"]["pout"].keys():
            node["meta"]["ports_map"]["pout"][port_name]["head"] = utilities.get_head_of_df_as_list(port_df, port_config, flow_id, node_key, port_name)
        return {"flow_id": flow_id, "node_key": node_key, "node": node}

if __name__ == "__main__":
    b = Builder()
    m = b.load_model()
    print(m)



import ast
import importlib
import json
import re
from contextlib import redirect_stderr, redirect_stdout
from io import StringIO

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Code:
    def exec(self, flow_id: str, node_key: str, pin, settings):
        script = []
        code_snippet: str = settings["code"] if "code" in settings else ""
        node_text: str = "CODE" if "code" in node_key.lower() else "SOURCE"
        script.append(f"\n# {node_text}")

        # Agrego los modulos y alias al entorno de variables globales
        used_modules = self.find_imports(code_snippet)
        for m in used_modules:
            try:
                if m["alias"]:
                    globals()[m["alias"]] = importlib.import_module(m["name"])
                else:
                    globals().update(importlib.import_module(m["name"]).__dict__)
            except ModuleNotFoundError as e:
                import os
                import subprocess

                current_path = os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
                )
                python_path = os.path.join(current_path, "python", "python")
                if os.path.exists(python_path):
                    subprocess.run(
                        [
                            python_path,
                            "-m",
                            "pipenv",
                            "install",
                            "--skip-lock",
                            m["name"].split(".")[0],
                        ]
                    )
                else:
                    import pip

                    pip.main(["install", m["name"].split(".")[0]])

                if m["alias"]:
                    globals()[m["alias"]] = importlib.import_module(m["name"])
                else:
                    globals().update(importlib.import_module(m["name"]).__dict__)

        def vtg_codeout(x):
            global vtg_df, vtg_metacode
            vtg_df = x.copy()

        loc = {f"df_{k.lower()}": v.copy() for k, v in pin.items()}
        loc["vtg_codeout"] = vtg_codeout
        stdout = ""
        # stderr = ""
        try:
            out = StringIO()
            # err = StringIO()
            # with redirect_stdout(out), redirect_stderr(err):
            with redirect_stdout(out):
                # NOTE: Antes estaba como exec(code_snippet, None, loc).
                # NOTE: Tener en cuenta en caso de que se encuentre algun fallo
                exec(code_snippet, loc)
            stdout = out.getvalue()
        except Exception as e:
            # stderr = err.getvalue()
            # print("err", stderr)
            msg = f"({node_text}) Exception:" + str(e)
            default = bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            default["STDOUT"] = stdout
            default["STDOUT"] += str(e)
            return default
        else:
            script.append(code_snippet)
            df_out = globals()["vtg_df"] if "vtg_df" in globals() else []
            globals()["vtg_df"] = []
            if not isinstance(df_out, pd.DataFrame):
                msg = f"({node_text}): Debes llamar a la función vtg_codeout(DF) con tu DataFrame de salida"
                default = bug_handler.default_on_error(flow_id, node_key, msg, "", "error")
                default["STDOUT"] = stdout
                return default

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df_out},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script

        return {"Out": df_out, "STDOUT": stdout}

    def find_imports(self, code_snippet):
        # Busco los modulos que tienen la forma ['import * as *']
        matchs = re.findall("import (.*?) as (.*?)$", code_snippet, flags=re.MULTILINE)
        # print(matchs)
        out = [{"name": m[0].strip(), "alias": m[1].strip(), "objects": []} for m in matchs]
        # print(out)
        # Busco los ['from * import *', 'import *']
        # modules = []
        for node in ast.iter_child_nodes(ast.parse(code_snippet)):
            if isinstance(node, ast.ImportFrom):
                objects = [node.names[i].name for i in range(len(node.names))]
                if not node.names[0].asname:  # excluding the 'as' part of import
                    # modules.append(node.module)
                    out.append({"name": node.module, "alias": None, "objects": objects})
            elif isinstance(node, ast.Import):  # excluding the 'as' part of import
                if not node.names[0].asname:
                    out.append({"name": node.names[0].name, "alias": None, "objects": []})
                    # modules.append(node.names[0].name)
        return out



import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.script_handler import script_handler


class Output:
    def exec(self, flow_id, node_key, pin, settings):
        script_handler.script.append("\n# OUPUT")
        output_format: str = settings["format"] or "csv"
        output_name: str = settings["name"] if "name" in settings else "output"
        output_path: str = settings["path"] if "path" in settings else ""
        encoding: str = settings["encoding"] or "UTF-8"
        # print(columns, order)
        df: pd.DataFrame = pin["In"].copy()
        try:
            if output_format == "excel":
                df.to_excel(
                    output_path + "\\" + output_name + ".xlsx",
                    index=False,
                    encoding=encoding,
                )
                script_handler.script.append(
                    "df.to_excel('{}.xlsx', index= False, encoding={})".format(
                        output_path + "\\" + output_name, encoding
                    )
                )
            else:
                df.to_csv(
                    output_path + "\\" + output_name + ".csv",
                    encoding=encoding,
                    index=False,
                )
                script_handler.script.append(
                    "df.to_csv('{}.csv', index= False, encoding={})".format(
                        output_path + "\\" + output_name, encoding
                    )
                )

        except Exception as e:
            print("Error (output): ", e)
            msg = f"(output) No fue posible aplicar el output: {str(e)}"
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        return df



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class IsIn:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info')
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        script_handler.script.append("\n# IS_IN")

        df_S: pd.DataFrame = pin["Src"].copy()
        df_V: pd.DataFrame = pin["Val"].copy()
        source: str = settings["source"]
        values: str = settings["values"]

        try:
            if "not_in" in settings and settings["not_in"]:
                df = df_S[~df_S[source].isin(df_V[values])].reset_index(drop=True)
            else:
                df = df_S[df_S[source].isin(df_V[values])].reset_index(drop=True)
            script_handler.script.append(
                "df = df[df_S[{}].isin(df_V[{}])].reset_index(drop=True)".format(source, values)
            )
        except Exception as e:
            msg = "(isin) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json
import os

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class InputData:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        # print(node_key)
        encoding: str = (
            settings["encoding"]
            if ("encoding" in settings and settings["encoding"] is not None)
            else "ISO-8859-1"
        )
        dtype = str if "as_string" in settings and settings["as_string"] == True else None
        delimiter: str = (
            settings["delimiter"]
            if "delimiter" in settings and settings["delimiter"] is not None
            else None
        )
        header: str = (
            None if "has_header" in settings and settings["has_header"] == False else "infer"
        )
        # print(encoding)
        # df = pd.read_csv(settings['file_path'], dtype=dtype, encoding=settings['encoding'])

        _, file_ext = os.path.splitext(settings["file_path"])
        file_ext = file_ext[1:]
        # print(filename, file_ext)
        try:
            bug_handler.console(
                'Leyendo fuente "{}"...'.format(settings["file_path"]), "trace", flow_id
            )
            if file_ext in ["csv", "txt"]:
                df = pd.read_csv(
                    settings["file_path"],
                    dtype=dtype,
                    encoding=encoding,
                    delimiter=delimiter,
                    header=header,
                    prefix="col_" if header is None else None,
                )
            elif file_ext == "json":
                orient = settings["orient"] if "orient" in settings else "columns"
                df = pd.read_json(settings["file_path"], orient=orient, encoding=encoding)
            elif file_ext in ["xls", "xlsx", "xlsm", "xlsb"]:
                sheet_name = settings["sheet_name"] if "sheet_name" in settings else 0
                df = pd.read_excel(settings["file_path"], dtype=dtype, sheet_name=sheet_name)
                # print(sheet_name)
            else:
                msg = f"(input): Formato {file_ext} no reconocido"
                bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

            df.columns = [str(c) for c in df.columns]

            # revisar si alguna nombre de columna tiene espacio al inicio o al final
            if True in [c.startswith((" ", "\t")) or c.endswith((" ", "\t")) for c in df.columns]:
                df.columns = [c.strip() for c in df.columns]
                msg = f"(input): Archivo fuente contiene espacios al inicio o final del nombre de una o más columnas. Se ha corregido para la lectura"
                bug_handler.default_on_error(
                    flow_id, node_key, msg, console_level="warn", bug_level="warning"
                )

        except Exception as e:
            msg = "(input) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        # TODO: Modificar por tipo de archivo, csv, json, excel
        script.append("\n# INPUT")
        script.append(
            "df = pd.read_csv('{}', encoding='{}')".format(settings["file_path"], encoding)
        )
        script.append(f"df = df.astype(str)")
        # script.append(f"df_{node_key}_Out = df_{node_key}_Out.astype(str)")

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import pandas as pd
import numpy as np
import urllib
import json

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


# Database Write
class DatabaseWrite:
    def exec(self, flow_id, node_key, pin, settings):
        # import snowflake.connector
        # from google.cloud import bigquery
        # from google.oauth2 import service_account
        # from pymongo import MongoClient
        from sqlalchemy import create_engine
        import pyodbc

        df: pd.DataFrame = pin["In"].copy()
        script = []
        script.append("\n# DATABASE WRITE")

        if_exist = settings["if_exist"] if ("if_exist" in settings and settings["if_exist"] is not None) else ''
        host = settings["host"] if ("host" in settings and settings["host"] is not None) else None
        password = ':'+settings["password"] if ("password" in settings and settings["password"] is not None) else ''
        port = ':'+str(settings["port"]) if ("port" in settings and settings["port"] is not None) else ''
        user = settings["user"] if ("user" in settings and settings["user"] is not None) else None
        project = ( settings["project"] if ("project" in settings and settings["project"] is not None) else None)
        database = ( settings["database"] if ("database" in settings and settings["database"] is not None) else None)
        table = (settings["table"] if ("table" in settings and settings["table"] is not None) else None)
        source = ( settings["source"] if ("source" in settings and settings["source"] is not None) else None )
        # type_connection = ( settings["type_connection"] if ("type_connection" in settings and settings["type_connection"] is not None) else None )

        # # Validación principal
        # if not type_connection or not source:
        #     msg = "(database_write) Existen campos vacíos"
        #     return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        # if type_connection == "sql_database" and (
        #     not host or not user or not password or not database or not table
        # ):
        #     msg = "(database_write) Existen campos vacíos"
        #     return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        # if type_connection == "cloud_database":
        #     if source == "bigquery" and (not host or not project or not database or not table):
        #         msg = "(database_write) Existen campos vacíos"
        #         return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
        #     elif source == "snowflake" and (
        #         not host or not user or not password or not project or not database
        #     ):
        #         msg = "(database_write) Existen campos vacíos"
        #         return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        # if type_connection == "nosql_database" and (not host or not database or not table):
        #     msg = "(database_write) Existen campos vacíos"
        #     return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        script.append("\n# DATABASE WRITE")
        if source == 'postgresql' or source == 'mysql' or source == 'sqlite' or source == 'oracle' or source == 'mariadb':
            try:
                # Se validan los campos de entrada
                if not host or not source or not database or not table or not if_exist:
                    msg = '(database_write) Existen campos vacíos'
                    return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
                
                connection = self.get_url_connection(source, user, password, host, port, database)
                engine = create_engine(connection)
                df.to_sql(name=table, con=engine, if_exists=if_exist, index=False)
                engine.dispose()
            except Exception as e:
                msg = "(database_write) Exception:" + str(e)
                return bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            
        elif source == 'sqlserver_2000':
            try:
                connection = pyodbc.connect("DRIVER={SQL Server};"+ f"User={user};Password={password};Database={database};Server={host};Port={port};")
                cursor = connection.cursor()

                # Preparación de datos
                columns_name = ', '.join(df.columns)
                values = ', '.join(['?' for x in df.columns])
                params = iter(np.asarray(df).tolist())

                # Limpia o no la tabla seleccionada de la base de datos
                if if_exist == 'replace': cursor.execute(f'TRUNCATE TABLE {table}')

                # Inserción
                cursor.executemany(f'INSERT INTO {table} ({columns_name}) VALUES ({values})', params)
                cursor.commit()
            except Exception as e:
                cursor.rollback()
                msg = "(database_write) Exception:" + str(e)
                return bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            
            cursor.close()
            connection.close()

        # elif source == "sqlserver_2000":
        #     # ['SQL Server', 'SQL Server Native Client 11.0', 'ODBC Driver 17 for SQL Server']
        #     connection = pyodbc.connect(
        #         "DRIVER={SQL Server};"
        #         + f"User={user};Password={password};Database={database};Server={host};Port={port};"
        #     )
        #     df.to_sql(schema=database, name=table, con=connection, if_exists=if_exist, index=False)
        #     connection.close()

        # elif source == "bigquery":
        #     with open(host) as file:
        #         host = json.load(file)
        #     credentials = service_account.Credentials.from_service_account_info(host)
        #     client = bigquery.Client(credentials=credentials)
        #     table_ref = client.dataset(database, project=project).table(table)
        #     rows = client.list_rows(table_ref)
        #     df = rows.to_dataframe()
        #     client.close()

        # elif source == "snowflake":
        #     connection = snowflake.connector.connect(
        #         user=user,
        #         password=password,
        #         account=host,
        #         database=project,
        #         schema=database,
        #     )

        #     query = f'SELECT * FROM "{table}"'
        #     cursor = connection.cursor()
        #     cursor.execute(query)
        #     results = cursor.fetchall()
        #     column_names = [desc[0] for desc in cursor.description]
        #     df = pd.DataFrame(results, columns=column_names)
        #     connection.close()
        #     cursor.close()

        # elif source == "mongodb":
        #     client = MongoClient(host)
        #     db = client[database]
        #     collection = db[table]
        #     data = list(collection.find())
        #     df = pd.DataFrame(data)
        #     client.close()
        else:
            msg = "(database_write) El tipo de conexión no coincide con ninguno"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}

    def get_url_connection(self, source, user, password, host, port, database):
        # if source == 'sqlserver_2000':
        #     quoted = urllib.parse.quote_plus('DRIVER={SQL Server};'+f'User={user};Password={password};Server={host};Port={port};Database={database}')    
        #     return 'mssql+pyodbc:///?odbc_connect={}'.format(quoted)
        if source == 'sqlite':
            # return f"sqlite:///:memory:"
            return f"sqlite:///{host}"
        elif source == 'mysql':
            return f"mysql+pymysql://{user}{password}@{host}{port}/{database}"
        elif source == 'postgresql':
            return f"postgresql://{user}{password}@{host}{port}/{database}"


# # -*- coding: utf-8 -*-

import io
import os
import smtplib
from email import encoders
from email.header import Header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Dict, List

"""
Envia un correo automático utilizando un App Password ded Google
https://towardsdatascience.com/automate-sending-emails-with-gmail-in-python-449cc0c3c317
"""


class EmailSender(object):
    def __init__(self):
        self.SERVER: str = None
        self.PORT: int = None
        self.FROM: str = None
        self.PASS: str = None
        self.TO: str = None
        self.SUBJECT: str = None
        self.HTML: str = None
        self.FILES: list[str] = []
        self.FOOTER: str = None
        self.prepare_footer()

    def configure_server(self, SERVER: str, PORT: int, FROM: str, PASS: str):
        self.FROM = FROM
        self.PASS = PASS
        self.SERVER = SERVER
        self.PORT = PORT

    def send_email(
        self, to: str, subject: str, message: str, files: List = None, streams: List[Dict] = None
    ):
        msg = MIMEMultipart()
        if to is not None:
            self.TO = to
        if subject is not None:
            self.SUBJECT = subject
        msg["From"] = self.FROM
        msg["To"] = self.TO
        msg["Subject"] = Header(self.SUBJECT)

        self.HTML = message

        if self.HTML:
            msg.attach(MIMEText(self.HTML, "html"))

        msg.attach(MIMEText(self.FOOTER, "html"))

        if files is not None and len(files):
            self.prepare_files(files)

        for f in self.FILES:
            part = MIMEBase("application", "octet-stream")
            part.set_payload(open(f, "rb").read())
            encoders.encode_base64(part)
            part.add_header(
                "Content-Disposition",
                'attachment; filename="{0}"'.format(os.path.basename(f)),
            )
            msg.attach(part)

        if streams is not None and len(streams):
            for f in streams:
                attach: io.StringIO = f["attachment"]
                if type(attach) is io.StringIO:
                    attachment = MIMEText(attach.getvalue())
                    attachment.add_header(
                        "Content-Disposition", "attachment", filename=f["filename"] or "out"
                    )
                    msg.attach(attachment)

        # image_path = f"{os.getcwd()}\\lib\\utils\\logo.png"

        # fp = open(image_path, 'rb')
        # msgImage = MIMEImage(fp.read())
        # fp.close()

        # # Define the image's ID as referenced above
        # msgImage.add_header('Content-ID', '<image1>')
        # msg.attach(msgImage)

        s = smtplib.SMTP(self.SERVER, self.PORT)
        s.ehlo()  # Hostname to send for this command defaults to the fully qualified domain name of the local host.
        s.starttls()  # Puts connection to SMTP server in TLS mode
        s.login(self.FROM, self.PASS)
        s.sendmail(self.FROM, self.TO.split(","), msg.as_string())
        s.quit()

    def prepare_footer(self):
        self.FOOTER = (
            "<br/><br/>"
            + "<p>This email was sent automatically</p>"
            +
            # "<img src=\"cid:image1\" alt=\"Logo\" style=\"height:70px;\"><br/>"
            '<p><a href="https://ai/">VTarget</a></p>'
        )

    def prepare_files(self, files=[]):
        self.FILES = list(filter(lambda x: os.path.exists(x), files))


# if __name__ == "__main__":
#     se = EmailSender()
#     se.send_email()



import json

import numpy as np
import pandas as pd

from vtarget.handlers.cache_handler import cache_handler


class Utilities:
    # retorna la metadata del dtypes de un df
    def get_dtypes_of_df(self, df: pd.DataFrame):
        dict_dtypes = {}
        res = df.dtypes.to_frame("dtypes")
        # print(res)
        res = res["dtypes"].astype(str).reset_index()
        res["selected"] = True
        res["order"] = pd.RangeIndex(stop=res.shape[0])
        # return res.values.tolist()
        # print(res.columns)
        for _, x in res.iterrows():
            dict_dtypes[x["index"]] = {
                "dtype": x["dtypes"],
                "selected": x["selected"],
                "order": x["order"],
            }

        return dict_dtypes
    
    def get_table_config(self, meta: dict, port_name: str):
        #? sacar config desde meta.ports_config con el puerto
        ports_config : dict = meta["ports_config"][port_name] if "ports_config" in meta and port_name in meta["ports_config"] else {}
        # print(ports_config)
        return {
            "rows": ports_config["rows"] if "rows" in ports_config and ports_config["rows"] is not None else 50,
            "decimals": ports_config["decimals"] if "decimals" in ports_config and ports_config["decimals"] is not None else -1,
            "source": ports_config["source"] if "source" in ports_config and ports_config["source"] else "head",
            "sort_by": ports_config["sort_by"] if "sort_by" in ports_config and ports_config["sort_by"] else [],
        }

    def sort_df(self, full_df: pd.DataFrame, sorts: list[dict]):
        if sorts:
            setting_list = list(map(lambda x: (x["field"], int(x["ascending"])), [item for item in sorts if "field" in item and item["field"]]))
            if setting_list:
                columns, order = zip(*setting_list)
                if columns and order:
                    full_df = full_df.sort_values(by=list(columns), ascending=list(order))
        return full_df
        
    def get_head_of_df_as_list(self, port_df: pd.DataFrame, config: dict, flow_id: str = None, node_key: str = None, port_name: str = None):
        df: pd.DataFrame = port_df.head(50).copy()
        rows = len(port_df) if config["rows"] > len(port_df) else config["rows"]
        
        # sort
        if "sort_by" in config and len(config["sort_by"]) > 0:
            port_df = self.sort_df(port_df, config["sort_by"])
            cached_node = cache_handler.get_node(flow_id, node_key)
            # actualizar cache del nodo con el nuevo order
            if flow_id and node_key and port_name:
                # actualizar cache del nodo con el nuevo order
                if cached_node["pout"]:
                    cached_node["pout"][port_name] = port_df
                    cache_handler.update_node(flow_id, node_key, { "pout": cached_node["pout"] })
                    
                # cached_node = cache_handler.get_node(flow_id, node_key)
                # if cached_node["pout"]:
                #     pout: dict = dict(cached_node["pout"]).update({ port_name: full_df })
                #     cache_handler.update_node(
                #         flow_id,
                #         node_key,
                #         { "pout": pout },
                #     )
        
        if config["source"] == "head":
            df = port_df[:rows].copy()
        elif config["source"] == "tail":
            df = port_df[-rows:].copy()
        elif config["source"] == "sample":
            df = port_df.sample(rows).copy()
            if "sort_by" in config and len(config["sort_by"]) > 0:
                df = self.sort_df(df, config["sort_by"])
        else:
            df = port_df.head(50).copy()
            print(
                "Source {} no reconocido opciones válidas [head|sample|tail]. Se utilizará head(50)".format(
                    config["source"]
                )
            )
            
        # decimals
        if config["decimals"] != -1:
            df = df.round(config["decimals"])
            
        # Esto para efectos de la visualización al transformar a json
        # special_cols = df.select_dtypes(include=['bool', 'datetime64', 'category']).columns.values.tolist()
        special_cols = df.select_dtypes(
            exclude=[
                "object",
                "int8",
                "int16",
                "int32",
                "int64",
                "float16",
                "float32",
                "float64",
            ]
        ).columns.values.tolist()
        
        if len(special_cols):
            # print(special_cols)
            df[special_cols] = df[special_cols].astype(str)
            
        df = df.fillna("NaN")
        # df = df.replace([np.inf, -np.inf], 0, inplace=True)
        # print(bool_cols)
        # df_head = [df.columns.values.tolist()] + df.values.tolist()
        df_head = df.to_dict("records")
        return df_head

    def format_setting(self, settings, ignore_keys=["ports_map", "readed_from_cache"]):
        setting_copy = {key: value for key, value in settings.items() if key not in ignore_keys}
        return json.dumps(setting_copy, sort_keys=True)

    def viz_summary(self, df):
        cat_col = df.select_dtypes(
            include=["object", "category", "bool", "datetime64", "timedelta"]
        ).columns.tolist()
        num_col = df.select_dtypes(
            include=["int16", "int32", "int64", "float16", "float32", "float64"]
        ).columns.tolist()
        # date_col = df.select_dtypes(include=['datetime64', 'timedelta']).columns.tolist()
        out = {}
        for c in num_col:
            count, bin_ = np.histogram(df[c][np.isfinite(df[c])])
            out[c] = {
                "viz_type": "histogram",
                "y": count.tolist(),
                "x": np.around(bin_, decimals=2).tolist(),
            }

        max_cat = 3
        for c in cat_col:
            cat_viz = "pie"
            vc = df[c].value_counts().iloc[:max_cat]
            vc.index = vc.index.astype("str")
            cat_counts = vc.to_dict()
            if df[c].nunique() > max_cat:
                others = df[~df[c].isin(vc.index)][c].value_counts()
                cat_counts[f"Other ({len(others)})"] = others.sum().item()
                cat_viz = "list"
            out[c] = {"viz_type": cat_viz, "values": cat_counts}
        return out

    def get_central_tendency_measures(self, df):
        if df.empty:
            return {}
        info = df.describe(include="all", datetime_is_numeric=True).T.reset_index()
        # info = info.astype(str)
        # print(info.dtypes)
        jsonlist = json.loads(info.to_json(orient="records"))
        return dict([(x["index"], x) for x in jsonlist])


utilities = Utilities()



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Unique:
    def exec(self, flow_id, node_key, pin, settings):
        script = []

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# Unique")

        # field
        field: list = (
            settings["field"] if ("field" in settings and settings["field"] is not None) else []
        )
        # print(field)
        if not field:
            msg = "(unique) Debes seleccionar al menos una columna para aplicar la función unique"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        try:
            df = df.groupby(field).size().reset_index().drop(columns=[0])
            script.append(
                "df = df.groupby({}).size().reset_index().drop(columns=[0])".format(field)
            )
            # new_df = pd.DataFrame()
            # script.append('new_df = pd.DataFrame()')
            # new_df[field] = df[field].unique()
            # script.append('new_df[{}] = df[{}].unique()'.format(field, field))
        except Exception as e:
            msg = "(unique) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Merge:
    def __init__(self):
        self.script = []

    def exec(self, flow_id, node_key, pin, settings):
        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs
        if "iL" not in pin or "iR" not in pin:
            msg = "(merge) Puerto entrada iL/iR no conectado"
            bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")
            edf = pd.DataFrame()
            return {"L": edf, "J": edf, "R": edf, "F": edf}

        self.script.append("\n# MERGE")
        df_iL: pd.DataFrame = pin["iL"][settings["left_columns"]].copy()
        df_iR: pd.DataFrame = pin["iR"][settings["right_columns"]].copy()

        on_l: list[str] = [x["left"] for x in settings["items"] if x["left"]]
        on_r: list[str] = [x["right"] for x in settings["items"] if x["right"]]

        # if len(settings['outputs']) == 4 or 'F' in settings['outputs']: # [L,J,R,F] o si incluye F
        # 	pout_struct = self.full_outer_join(flow_id, node_key, df_iL, df_iR, on_l, on_r)
        # else:
        edf = pd.DataFrame()
        pout_struct = {"L": edf, "J": edf, "R": edf, "F": edf}
        if "L" in settings["outputs"]:
            pout_struct["L"] = self.apply_join(flow_id, node_key, "left", df_iL, df_iR, on_l, on_r)
        if "J" in settings["outputs"]:
            pout_struct["J"] = self.apply_join(flow_id, node_key, "inner", df_iL, df_iR, on_l, on_r)
        if "R" in settings["outputs"]:
            pout_struct["R"] = self.apply_join(flow_id, node_key, "right", df_iL, df_iR, on_l, on_r)
        if "F" in settings["outputs"]:
            pout_struct["F"] = self.apply_join(flow_id, node_key, "outer", df_iL, df_iR, on_l, on_r)

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": pout_struct,
                "config": json.dumps(settings, sort_keys=True),
                "script": self.script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += self.script
        return pout_struct

    def apply_join(self, flow_id, node_key, how, df_iL, df_iR, on_l, on_r):
        try:
            if how == "outer":
                df_o = pd.merge(
                    df_iL,
                    df_iR,
                    left_on=on_l,
                    right_on=on_r,
                    how="outer",
                    indicator=True,
                )  # , validate="many_to_one")
                # campo _merge viene como category, lo cambio a object (str)
                df_o["merge_type"] = df_o["_merge"].astype(str)
                del df_o["_merge"]
                # df_o["_merge"] = df_o["_merge"].astype(str)
                self.script.append(
                    "df_o = pd.merge(df_iL, df_iR, left_on={}, right_on={}, how='outer', indicator=True)".format(
                        on_l, on_r
                    )
                )
            else:
                # , validate="many_to_one")
                df_o = pd.merge(df_iL, df_iR, left_on=on_l, right_on=on_r, how=how)
                self.script.append(
                    "df_o = pd.merge(df_iL, df_iR, left_on={}, right_on={}, how='{}', indicator=True)".format(
                        on_l, on_r, how
                    )
                )
        except Exception as e:
            msg = "(merge) Exception:" + str(e)
            bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            return pd.DataFrame()
        return df_o

    # def full_outer_join(self, flow_id, node_key, df_iL, df_iR, on_l, on_r):
    # 	try:
    # 		# Outer join con indicator (que genera la columna merge) para sacar sólo las columnas de lado correspondiente
    # 		df_o = pd.merge(df_iL, df_iR, left_on=on_l, right_on=on_r, how="outer", indicator=True)#, validate="many_to_one")
    # 		self.script.append("df_o = pd.merge(df_iL, df_iR, left_on={}, right_on={}, how='outer', indicator=True)".format(on_l, on_r))
    # 	except Exception as e:
    # 		msg = '(merge) Exception:' + str(e)
    # 		bug_handler.console(msg, 'fatal', flow_id)
    # 		bug_handler.append({'flow_id': flow_id, 'success': False, 'node_key': node_key, 'level': 'error', 'msg': msg, 'exception': str(e)})
    # 		edf = pd.DataFrame()
    # 		return {'L': edf, 'J': edf, 'R': edf, 'F': edf}

    # 	# print('J',df_iL.shape, df_iR.shape)
    # 	df_l = df_o[df_o['_merge'].isin(['left_only', 'both'])].drop(columns=['_merge'], axis=1) # where 1 is the axis number (0 for rows and 1 for columns.)
    # 	df_j = df_o[df_o['_merge'] == 'both'].drop(columns=['_merge'], axis=1) # where 1 is the axis number (0 for rows and 1 for columns.)
    # 	df_r = df_o[df_o['_merge'].isin(['right_only', 'both'])].drop(columns=['_merge'], axis=1) # where 1 is the axis number (0 for rows and 1 for columns.)

    # 	self.script.append("df_l = df_o[df_o['_merge'] == 'left_only'].drop(columns=['_merge'], axis=1)")
    # 	self.script.append("df_j = df_o[df_o['_merge'] == 'both'].drop(columns=['_merge'], axis=1)")
    # 	self.script.append("df_r = df_o[df_o['_merge'] == 'right_only'].drop(columns=['_merge'], axis=1)")

    # 	df_o['_merge'] = df_o['_merge'].astype(str) # campo _merge viene como category, lo cambio a object (str)
    # 	# df_o.rename(columns={'_merge': 'merge'}, inplace=True)
    # 	# print(df_j.head())

    # 	return {'L': df_l, 'J': df_j, 'R': df_r, 'F': df_o}



from enum import Enum


class NodeType(Enum):
    BARCHART = "Bar_Chart"
    CODE = "Code"
    COLUMN = "Column"
    COMMENT = "Comment"
    CONCAT = "Concat"
    CROSSJOIN = "Cross_Join"
    CUMSUM = "Cumsum"
    DATABASE = "Database"
    DATACLEANSING = "Data_Cleansing"
    DATETIMEEXTRACT = "Datetime_Extract"
    DATETIMEFORMATTER = "Datetime_Formatter"
    DESCRIBE = "Describe"
    DROPDUPLICATES = "Drop_Duplicates"
    DTYPE = "Dtype"
    FILTER = "Filter"
    FORMULA = "Formula"
    GROUPBY = "Group_By"
    INPUTDATA = "Input_Data"
    INTERROW = "Inter_Row"
    ISIN = "Isin"
    JOINMULTI = "Join_Multi"
    LINECHART = "Line_Chart"
    MELT = "Melt"
    MERGE = "Merge"
    OUTPUTDATA = "Output_Data"
    PIECHART = "Pie_Chart"
    PIVOT = "Pivot"
    SCATTERCHART = "Scatter_Chart"
    SELECT = "Select"
    SHAPE = "Shape"
    SORT = "Sort"
    SWITCH = "Switch"
    UNIQUE = "Unique"
    VALUECOUNTS = "Value_Counts"
    VOUTPUT = "V_Output"
    EXCEL = "Excel"
    EMAIL = "Email"
    IF = "If"
    WAIT = "Wait"
    DFMAKER = "DF_Maker"
    SOURCE = "Source"



from setuptools import __version__, setup

with open("./README.md", "r") as f:
    readme = f.read()

with open("./requirements.txt", "r") as f:
    requirements = f.read().splitlines()

setup(
    name="vtarget",
    packages=[
        "vtarget",
        "vtarget.dataprep",
        "vtarget.dataprep.nodes",
        "vtarget.handlers",
        "vtarget.utils",
    ],
    version="0.1.8",
    description="vtarget lib",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="vTarget Team",
    author_email="contact@vtarget.ai",
    keywords=["vtarget", "dataprep"],
    classifiers=[],
    license="BSD",
    install_requires=requirements,
    include_package_data=False,
)



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler


class Describe:
    def exec(self, flow_id, node_key, pin, settings):
        script = []
        # if node_key in cache_handler.cache[flow_id] and cache_handler.cache[flow_id][node_key]['config'] == json.dumps(settings, sort_keys=True):
        # 	bug_handler.console(f'Nodo "{node_key}" leído desde cache flow_id: "{flow_id}"', 'info', flow_id)
        # 	reset_childs = False
        # 	script_handler.script += cache_handler.cache[flow_id][node_key]['script']
        # 	return cache_handler.cache[flow_id][node_key]['pout'], reset_childs

        df: pd.DataFrame = pin["In"].copy()
        script.append("\n# DESCRIBE")

        all_: str = "all"  # TODO: recibir desde la config
        groupby: list = (
            settings["groupby"] if ("groupby" in settings and settings["groupby"] is not []) else []
        )
        fields: list = (
            settings["fields"] if ("fields" in settings and settings["fields"] is not None) else []
        )
        percentiles: list = (
            settings["percentiles"]
            if ("percentiles" in settings and settings["percentiles"] is not None)
            else []
        )
        custom_percentiles: list = (
            settings["custompercentiles"]
            if ("custompercentiles" in settings and settings["custompercentiles"] is not None)
            else []
        )

        if "pivot_data" not in settings:
            settings["pivot_data"] = True

        pivot_data: bool = (
            settings["pivot_data"]
            if ("pivot_data" in settings and settings["pivot_data"] is not None)
            else True
        )

        if not fields:
            msg = "(describe) Debes seleccionar al menos una columna para aplicar el describe"
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        fields = list(filter(lambda x: x in df.columns, fields))
        all_percentiles = percentiles + custom_percentiles
        # quitar percentiles repetidos
        if type(all_percentiles) is list and len(all_percentiles) > 0:
            all_percentiles = list(set(all_percentiles))

        try:
            if groupby:
                df = (
                    df.groupby(by=groupby)[fields]
                    .describe(include=all_, percentiles=all_percentiles)
                    .reset_index()
                )
                script.append(
                    "df = df.groupby(by={})[{}].describe(include='{}', percentiles={}).reset_index()".format(
                        groupby, fields, all_, all_percentiles
                    )
                )

                if len(fields) > 1:
                    df.columns = ["_".join(list(map(str, x))) if x[1] else x[0] for x in df.columns]
                    script.append(
                        "df.columns = ['_'.join(list(map(str, x))) if x[1] else x[0] for x in df.columns]"
                    )
                else:
                    df.columns = [x[1] if x[1] else x[0] for x in df.columns]
                    script.append("df.columns = [x[1] if x[1] else x[0] for x in df.columns]")
            else:
                df = df[fields].describe(include=all_, percentiles=all_percentiles)

                script.append(
                    "df = df.describe(include='{}', percentiles={})".format(all_, all_percentiles)
                )

                if pivot_data and len(fields) == 1:
                    df = df.T.reset_index().rename(columns={"index": "column"})
                    script.append("# pivot data")
                    script.append("df = df.T.reset_index().rename(columns={'index': 'column'})")
                else:
                    df = df.reset_index()
                    script.append("df=df.reset_index()")

        except Exception as e:
            msg = "(describe) Exception:" + str(e)
            return bug_handler.default_on_error(flow_id, node_key, msg, str(e))

        df = df.round(2)

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        script_handler.script += script
        return {"Out": df}



import io
import json
import os

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler
from vtarget.handlers.script_handler import script_handler
from vtarget.utils.email_sender import EmailSender
from vtarget.utils.encrypter import encrypter


class Email:
    def exec(self, flow_id, node_key, pin, settings):
        streams: list[dict] = []
        files: list[str] = []
        ports: list[str] = [p for p in dict(settings).keys() if str(p).startswith("port_")]
        total_size: int = 0

        for port in ports:
            port_origin: str = settings[port] if port in settings else None
            if port_origin:
                # * Extraer key y puerto de salida del nodo
                sepIdx = len(port_origin) - port_origin[::-1].find("_")
                parent_key: str = port_origin[0 : sepIdx - 1]
                parent_port: str = port_origin[sepIdx:]

                if parent_key is not None and parent_port is not None:
                    # ? obtener el df del padre desde cache
                    parent = cache_handler.get_node(flow_id, node_key=parent_key)

                    if parent_key.lower().startswith("excel"):  # * Nodos de tipo excel
                        # ? obtener la ruta del archivo excel generado
                        parent_config: dict = (
                            json.loads(parent["config"]) if "config" in parent else {}
                        )
                        file_path: str = (
                            parent_config["file_path"] if "file_path" in parent_config else None
                        )
                        if file_path is None:
                            output_name = parent_config["name"] if "name" in parent_config else None
                            output_path = parent_config["path"] if "path" in parent_config else None
                            if output_name and output_path:
                                file_path = f"{output_path}\\{output_name}.xlsx"

                        # ? agregar la ruta del archivo a la lista de archivos
                        if file_path is not None:
                            files.append(file_path)

                            # ? sumar peso del excel al total
                            size = os.path.getsize(file_path)
                            total_size += size

                    else:  # * cualquier otro tipo de nodos
                        # ? convertir df del puerto de origen en un csv en memoria
                        df_in: pd.DataFrame = (
                            parent["pout"][parent_port]
                            if "pout" in parent and parent_port in parent["pout"]
                            else pd.DataFrame()
                        )
                        if len(df_in):
                            stream = io.StringIO()
                            df_in.to_csv(stream, index=False)

                            # ? sumar peso del csv al total
                            pos = stream.tell()
                            stream.seek(0, os.SEEK_END)
                            stream.seek(pos)
                            total_size += pos

                            # ? agregar el stream a la lista y asignarle un nombre por defecto
                            streams.append({"filename": f"{port_origin}.csv", "attachment": stream})

        # * check if total size > 20Mb
        if total_size / (1024 * 1024) > 20:
            msg = "(email): No es posible enviar el correo ya que el tamaño de los archivos adjunto supera el límite máximo"
            return bug_handler.default_on_error(flow_id, node_key, msg)

        script_handler.script.append("\n# EMAIL")
        SERVER: str = (
            settings["server"] if "server" in settings and settings["server"] is not None else None
        )
        PORT: str = (
            settings["port"] if "port" in settings and settings["port"] is not None else None
        )
        FROM: str = (
            settings["from"] if "from" in settings and settings["from"] is not None else None
        )
        PASS: str = (
            settings["password"]
            if "password" in settings and settings["password"] is not None
            else None
        )
        to: list = settings["to"] if "to" in settings and settings["to"] is not None else []
        subject = settings["subject"] if "subject" in settings else "-"
        message = settings["message"] if "message" in settings else ""

        validation = list(
            set(["server", "port", "from", "password", "to"])
            - set([k for k in settings.keys() if settings[k]])
        )
        if len(validation) != 0:
            msg = (
                "(email): Debes completar todos los campos en la configuración. Campos faltantes: "
                + str(validation)
            )
            return bug_handler.default_on_error(flow_id, node_key, msg, console_level="error")

        decryptKey = flow_id[:16]
        DECRYPTEDPASS = (
            encrypter.decrypt(PASS, decryptKey)
            if PASS is not None and decryptKey is not None
            else ""
        )

        try:
            emailSender = EmailSender()
            emailSender.configure_server(SERVER, PORT, FROM, DECRYPTEDPASS)
            emailSender.send_email(",".join(to), subject, message, streams=streams, files=files)

        except Exception as e:
            msg = "(concat) No fue posible realizar el envío de email:" + str(e)
            bug_handler.default_on_error(flow_id, node_key, msg, str(e))
            return {}

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "config": json.dumps(settings, sort_keys=True),
                "script": script_handler.script,
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)

        return {}



import json

import pandas as pd

from vtarget.handlers.bug_handler import bug_handler
from vtarget.handlers.cache_handler import cache_handler


class VOutput:
    def exec(self, flow_id, node_key, pin, settings):
        df: pd.DataFrame = pin["In"].copy()

        cache_handler.update_node(
            flow_id,
            node_key,
            {
                "pout": {"Out": df},
                "config": json.dumps(settings, sort_keys=True),
                "script": [],
                "type": "V_Output",
            },
        )

        bug_handler.console(f'[Nodo]: "{node_key}" almacenado en cache', "info", flow_id)
        # script_handler.script += script

        return pd.DataFrame()
